import random


def discrete_sample(points, get_probability):
    """
    Given a discrete set {e1, ... } and a probability function p,
    returns the first element in the set such that u < sum_i p(e_i)
    where u is uniformly sampled from [0; 1].
    This has the effect of sampling from the discrete set according to the probability function.
    """
    u = random.uniform(0, 1)
    total_probability_so_far = 0
    for point in points:
        total_probability_so_far += get_probability(point)
        if u < total_probability_so_far:
            return point
    raise Exception(f"Discrete probability provided for sampling is not normalized")



from typing import Any

import torch
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)
from neuralpp.util.util import is_iterable


class IntegerVariable(DiscreteVariable):
    def __init__(self, name, cardinality):
        super().__init__(name, cardinality)

    def assignments(self):
        return range(self.cardinality)

    def featurize(self, integer_value) -> torch.Tensor:
        if self.is_multivalue(integer_value):
            return torch.tensor(integer_value, dtype=torch.float).unsqueeze(1)
        else:
            return torch.tensor([integer_value], dtype=torch.float)

    def is_multivalue(self, value: Any) -> bool:
        return is_iterable(value)

    def multivalue_len(self, value: Any) -> int:
        assert self.is_multivalue(value)
        return len(value)

    def __eq__(self, other):
        assert isinstance(
            other, IntegerVariable
        ), "IntegerVariable being compared to non-IntegerVariable"
        result = self.name == other.name
        if result and self.cardinality != other.cardinality:
            print(
                f"Warning: variables {self} and {other} have the same name but different"
                f"cardinalities {self.cardinality} and {other.cardinality}"
            )
        return result

    def __hash__(self):
        return hash(self.name) + 31 * self.cardinality

    def __repr__(self):
        return str(self.name)



import torch
from neuralpp.util.batch_argmax import NoArgMaxIndices, batch_argmax
from neuralpp.util.util import check_that_exception_is_thrown


def test_basic():
    # a simple array
    tensor = torch.tensor([0, 1, 2, 3, 4])
    batch_dim = 0
    expected = torch.tensor(4)
    run_test(tensor, batch_dim, expected)

    # making batch_dim = 1 renders the non-batch portion empty and argmax indices undefined
    tensor = torch.tensor([0, 1, 2, 3, 4])
    batch_dim = 1
    check_that_exception_is_thrown(
        lambda: batch_argmax(tensor, batch_dim), NoArgMaxIndices
    )

    # now a batch of arrays
    tensor = torch.tensor([[1, 2, 3], [6, 5, 4]])
    batch_dim = 1
    expected = torch.tensor([2, 0])
    run_test(tensor, batch_dim, expected)

    # Now we have an empty batch with non-batch 3-dim arrays' shape (the arrays are actually non-existent)
    tensor = torch.ones(0, 3)  # 'ones' is irrelevant since this is empty
    batch_dim = 1
    # empty batch of the right shape: just the batch dimension 0,since indices of arrays are scalar (0D)
    expected = torch.ones(0)
    run_test(tensor, batch_dim, expected)

    # Now we have an empty batch with non-batch matrices' shape (the matrices are actually non-existent)
    tensor = torch.ones(0, 3, 2)  # 'ones' is irrelevant since this is empty
    batch_dim = 1
    # empty batch of the right shape: the batch and two dimension for the indices since we have 2D matrices
    expected = torch.ones(0, 2)
    run_test(tensor, batch_dim, expected)

    # a batch of 2D matrices:
    tensor = torch.tensor([[[1, 2, 3], [6, 5, 4]], [[2, 3, 1], [4, 5, 6]]])
    batch_dim = 1
    expected = torch.tensor(
        [[1, 0], [1, 2]]
    )  # coordinates of two 6's, one in each 2D matrix
    run_test(tensor, batch_dim, expected)

    # same as before, but testing that batch_dim supports negative values
    tensor = torch.tensor([[[1, 2, 3], [6, 5, 4]], [[2, 3, 1], [4, 5, 6]]])
    batch_dim = -2
    expected = torch.tensor([[1, 0], [1, 2]])
    run_test(tensor, batch_dim, expected)

    # Same data, but a 2-dimensional batch of 1D arrays!
    tensor = torch.tensor([[[1, 2, 3], [6, 5, 4]], [[2, 3, 1], [4, 5, 6]]])
    batch_dim = 2
    expected = torch.tensor([[2, 0], [1, 2]])  # coordinates of 3, 6, 3, and 6
    run_test(tensor, batch_dim, expected)

    # same as before, but testing that batch_dim supports negative values
    tensor = torch.tensor([[[1, 2, 3], [6, 5, 4]], [[2, 3, 1], [4, 5, 6]]])
    batch_dim = -1
    expected = torch.tensor([[2, 0], [1, 2]])
    run_test(tensor, batch_dim, expected)


def run_test(tensor, batch_dim, expected):
    actual = batch_argmax(tensor, batch_dim)
    print(
        f"batch_argmax of {tensor} with batch_dim {batch_dim} is\n{actual}\nExpected:\n{expected}"
    )
    assert actual.shape == expected.shape
    assert actual.eq(expected).all()



from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frame,
    number_of_equal_values_in_dict_frames,
    to,
)
from neuralpp.inference.graphical_model.variable_elimination import VariableElimination
from neuralpp.util.util import mean


def cross_entropy_for_dataset(dataset, model, debug=False):
    return mean(
        cross_entropy_for_datapoint(observation, query_assignment, model, debug)
        for (observation, query_assignment) in dataset
    )


def cross_entropy_for_datapoint(observation_dict, query_dict, model, debug=False):
    probability = compute_query_probability(observation_dict, query_dict, model, debug)
    cross_entropy_loss = -probability.log().sum()
    return cross_entropy_loss


def compute_accuracy_on_frames_data_loader(data_loader, model, device):
    total_number_of_correct_predictions = 0
    total_number_of_predictions = 0
    for (observation_batch, query_batch) in data_loader:
        observation_batch = to(observation_batch, device)
        query_batch = to(query_batch, device)
        (
            number_of_correct_predictions,
            number_of_predictions,
        ) = compute_number_of_correct_and_total_predictions(
            observation_batch, query_batch, model
        )
        total_number_of_correct_predictions += number_of_correct_predictions
        total_number_of_predictions += number_of_predictions
    total_accuracy = total_number_of_correct_predictions / total_number_of_predictions
    return total_accuracy


def compute_number_of_correct_and_total_predictions(
    observation_dict, query_dict, model
):
    query_variables = query_dict.keys()
    prediction_dict = compute_query_prediction(observation_dict, query_variables, model)
    number_of_correct_predictions = number_of_equal_values_in_dict_frames(
        prediction_dict, query_dict
    )
    number_of_predictions = generalized_len_of_dict_frame(query_dict)
    return number_of_correct_predictions, number_of_predictions


def compute_query_prediction(observation_dict, query_variables, model):
    probability = compute_query_distribution(observation_dict, query_variables, model)
    prediction = probability.argmax()
    return prediction


def compute_query_probability(observation_dict, query_dict, model, debug=False):
    query_variables = query_dict.keys()
    query_distribution = compute_query_distribution(
        observation_dict, query_variables, model
    )
    probability = query_distribution[query_dict]

    if debug:
        # print(f"Observation: {observation_dict}")
        print(f"Target: {query_dict}")
        # print(f"Query distribution: {query_distribution}")
        print(f"Predicted probability of target: {probability}")

    return probability


def compute_query_distribution(observation_dict, query_variables, model):
    conditioned_model = [f.condition(observation_dict) for f in model]
    query_distribution = VariableElimination().run(query_variables, conditioned_model)
    return query_distribution



import time

import torch
from neuralpp.util.every_k_times import EveryKTimes
from neuralpp.util.util import join
from torch import autograd
from tqdm import tqdm


# def default_after_epoch(learner, end_str='\n'):
#     include_decrease = learner.previous_epoch_average_loss is not None
#     time_elapsed = time.time() - learner.time_start
#     end_line = ' ' if include_decrease else end_str
#     print(f"[{time_elapsed:.0f} s] Epoch average loss: {learner.epoch_average_loss:.4f}", end=end_line)
#     if include_decrease:
#         loss_decrease = learner.previous_epoch_average_loss - learner.epoch_average_loss
#         print(f"(decrease: {loss_decrease:.6f}; stopping value is {learner.loss_decrease_tol:.6f})", end=end_str)


def default_after_epoch(learner, end_str="\n"):
    time_elapsed = time.time() - learner.time_start
    main_info = (
        f"[{time_elapsed:.0f} s] Epoch average loss: {learner.epoch_average_loss:.4f}"
    )
    extra_items = []
    include_decrease = learner.previous_epoch_average_loss is not None
    if include_decrease:
        loss_decrease = learner.previous_epoch_average_loss - learner.epoch_average_loss
        extra_items.append(f"decrease: {loss_decrease:.6f}")
        extra_items.append(f"stopping value is {learner.loss_decrease_tol:.6f}")
    if learner.max_epochs_to_go_before_stopping_due_to_loss_decrease > 1:
        extra_items.append(
            f"number of epochs without less decrease: {learner.number_of_epochs_without_loss_decrease}"
        )
        extra_items.append(
            f"max number of epochs without less decrease: {learner.max_epochs_to_go_before_stopping_due_to_loss_decrease}"
        )
    final_message = (
        main_info + " (" + join(extra_items, "; ") + ")" if extra_items else main_info
    )
    print(final_message, end=end_str)


class GenericSGDLearner:
    """
    A generic implementation of Stochastic Gradient Descent that does not make many assumptions regarding the
    actual data and model being learned.

    Attributes available to all methods:
    model
    data_loader
    lr
    loss_decrease_tol
    number_of_batches_between_updates
    device
    debug
    max_epochs_to_go_before_stopping_due_to_loss_decrease

    optimizer
    detect_autograd_anomaly
    parameters
    retain_graph_during_backpropagation
    time_start
    epoch
    epoch_average_loss
    epoch_average_loss_decrease,
    total_epoch_loss
    total_number_of_data_points
    previous_epoch_average_loss
    """

    def __init__(
        self,
        model,
        data_loader,
        lr=1e-3,
        loss_decrease_tol=1e-3,
        device=None,
        number_of_batches_between_updates=1,
        debug=False,
        after_epoch=default_after_epoch,
        max_epochs_to_go_before_stopping_due_to_loss_decrease=1,
    ):

        self.model = model
        self.data_loader = data_loader
        self.lr = lr
        self.loss_decrease_tol = loss_decrease_tol
        self.device = device
        self.number_of_batches_between_updates = number_of_batches_between_updates
        self.debug = debug
        self.after_epoch = after_epoch
        self.max_epochs_to_go_before_stopping_due_to_loss_decrease = (
            max_epochs_to_go_before_stopping_due_to_loss_decrease
        )

        self.parameters = self.get_parameters(model)
        self.retain_graph_during_backpropagation = (
            self.must_retain_graph_during_backpropagation()
        )

        # Attributes to be initialized at learning time
        self.optimizer = None
        self.time_start = None
        self.epoch = None
        self.total_epoch_loss = None
        self.total_number_of_data_points = None
        self.epoch_average_loss = None
        self.previous_epoch_average_loss = None
        self.epoch_average_loss_decrease = None
        self.number_of_epochs_without_loss_decrease = None

    def batch_to_device(self, batch, device):
        """Default implementation assumes batch is a tensor and returns its copy on device"""
        return batch.to(device)

    def get_number_of_datapoints_in_batch(self, batch):
        return len(batch)

    def loss_function(self, batch):
        error = NotImplementedError(f"loss_function not implemented for {type(self)}")
        raise error

    def get_parameters(self, model):
        return model.parameters()

    def must_retain_graph_during_backpropagation(self):
        return False

    def learning_is_done(self):
        no_loss_decrease = (
            self.epoch_average_loss_decrease is not None
            and self.epoch_average_loss_decrease <= self.loss_decrease_tol
        )
        if no_loss_decrease:
            self.number_of_epochs_without_loss_decrease += 1
        else:
            self.number_of_epochs_without_loss_decrease = 0
        done = (
            self.number_of_epochs_without_loss_decrease
            == self.max_epochs_to_go_before_stopping_due_to_loss_decrease
        )
        return done

    def learn(self):

        self.optimizer = torch.optim.Adam(self.parameters, lr=self.lr)
        self.optimizer.zero_grad()
        self.time_start = time.time()
        self.epoch = 0
        self.total_epoch_loss = None
        self.total_number_of_data_points = None
        self.epoch_average_loss = None
        self.previous_epoch_average_loss = None
        self.epoch_average_loss_decrease = None
        self.number_of_epochs_without_loss_decrease = 0

        def update_parameters():
            self.optimizer.step()
            self.optimizer.zero_grad()

        update_parameters_every_k_batches = EveryKTimes(
            update_parameters, self.number_of_batches_between_updates
        )

        while not self.learning_is_done():
            print(f"Epoch: {self.epoch + 1}")
            self.total_epoch_loss = 0
            self.total_number_of_data_points = 0

            for batch_number, batch in tqdm(enumerate(self.data_loader, start=1)):
                batch = self.batch_to_device(batch, self.device)
                number_of_data_points = self.get_number_of_datapoints_in_batch(batch)
                batch_loss = self.loss_function(batch)
                batch_loss.backward(
                    retain_graph=self.retain_graph_during_backpropagation
                )
                update_parameters_every_k_batches()
                self.total_epoch_loss += batch_loss.item()
                self.total_number_of_data_points += number_of_data_points

            self.epoch_average_loss = (
                self.total_epoch_loss / self.total_number_of_data_points
            )

            if self.previous_epoch_average_loss is not None:
                self.epoch_average_loss_decrease = (
                    self.previous_epoch_average_loss - self.epoch_average_loss
                )

            with torch.no_grad():
                self.after_epoch(self)

            self.previous_epoch_average_loss = self.epoch_average_loss
            self.epoch += 1

        return self.epoch_average_loss



import os
import pickle

from neuralpp.util.util import check_path_like


def pickle_cache(thunk, path, refresh=False):
    if not refresh:
        check_path_like(path, caller="pickle_cache")
        try:
            with open(path, "rb") as file:
                print(f"Reading from {path}...")
                result = pickle.load(file)
                print(f"Finished reading from {path}...")
                return result
        except FileNotFoundError:
            pass

    print(f"Computing from scratch...")
    value = thunk()
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "wb") as file:
        print(f"Saving to {path}...")
        pickle.dump(value, file)
    return value



from math import prod
from typing import Tuple

import torch
from neuralpp.util.timer import Timer


def unravel_indices(
    indices: torch.LongTensor,
    shape: Tuple[int, ...],
) -> torch.LongTensor:
    r"""Converts flat indices into unraveled coordinates in a target shape.

    Args:
        indices: A tensor of (flat) indices, (*, N).
        shape: The targeted shape, (D,).

    Example:
        indices = torch.stack((torch.arange(27).long(), torch.arange(27).long())).long()
        print(unravel_indices(indices, (3, 3, 3)))
        print(unravel_indices(indices, (9, 3)))
        print(unravel_indices(indices, (9, 1, 3)))

    Returns:
        The unraveled coordinates, (*, N, D).
    """

    coord = []

    for dim in reversed(shape):
        coord.append(indices % dim)
        indices = indices // dim

    coord = torch.stack(coord[::-1], dim=-1).long()

    return coord


def batch_argmax(tensor, batch_dim=1):
    """
    Assumes that dimensions of tensor up to batch_dim are "batch dimensions"
    and returns the indices of the max element of each "batch row".
    More precisely, returns tensor `a` such that, for each index v of tensor.shape[:batch_dim], a[v] is
    the indices of the max element of tensor[v].
    """
    if batch_dim >= len(tensor.shape):
        raise NoArgMaxIndices()
    batch_shape = tensor.shape[:batch_dim]
    non_batch_shape = tensor.shape[batch_dim:]
    flat_non_batch_size = prod(non_batch_shape)
    tensor_with_flat_non_batch_portion = tensor.reshape(
        *batch_shape, flat_non_batch_size
    )

    dimension_of_indices = len(non_batch_shape)

    # We now have each batch row flattened in the last dimension of tensor_with_flat_non_batch_portion,
    # so we can invoke its argmax(dim=-1) method. However, that method throws an exception if the tensor
    # is empty. We cover that case first.
    if tensor_with_flat_non_batch_portion.numel() == 0:
        # If empty, either the batch dimensions or the non-batch dimensions are empty
        batch_size = prod(batch_shape)
        if batch_size == 0:  # if batch dimensions are empty
            # return empty tensor of appropriate shape
            batch_of_unraveled_indices = torch.ones(
                *batch_shape, dimension_of_indices
            ).long()  # 'ones' is irrelevant as it will be empty
        else:  # non-batch dimensions are empty, so argmax indices are undefined
            raise NoArgMaxIndices()
    else:  # We actually have elements to maximize, so we search for them
        indices_of_non_batch_portion = tensor_with_flat_non_batch_portion.argmax(dim=-1)
        batch_of_unraveled_indices = unravel_indices(
            indices_of_non_batch_portion, non_batch_shape
        )

    if dimension_of_indices == 1:
        # above function makes each unraveled index of a n-D tensor a n-long tensor
        # however indices of 1D tensors are typically represented by scalars, so we squeeze them in this case.
        batch_of_unraveled_indices = batch_of_unraveled_indices.squeeze(dim=-1)
    return batch_of_unraveled_indices


class NoArgMaxIndices(BaseException):
    def __init__(self):
        super(NoArgMaxIndices, self).__init__(
            "no argmax indices: batch_argmax requires non-batch shape to be non-empty"
        )



from neuralpp.inference.graphical_model.representation.factor.table_factor import (
    TableFactor,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_table import (
    PyTorchTable,
)
from neuralpp.inference.graphical_model.representation.table.table_util import shape


class PyTorchTableFactor(TableFactor):
    def __init__(
        self, variables, array_or_table_of_potentials, log_space=True, batch=False
    ):
        super().__init__(
            variables,
            array_or_table_of_potentials
            if isinstance(array_or_table_of_potentials, PyTorchTable)
            else PyTorchLogTable.from_array(array_or_table_of_potentials, batch=batch)
            if log_space
            else PyTorchTable.from_array(array_or_table_of_potentials, batch=batch),
        )
        assert self.table.non_batch_shape == (
            variables_shape := tuple(v.cardinality for v in variables)
        ), (
            f"{type(self).__name__} created with variables {variables} with shape {variables_shape} "
            f"but with table with (non-batch) shape {self.table.non_batch_shape}"
        )

    def pytorch_parameters(self):
        return self.table.pytorch_parameters()

    @staticmethod
    def from_function_and_constructor(
        variables, function, constructor, log_space=True, batch_size=None
    ):
        """
        Builds a PyTorchTableFactor from neuralpp.its variables, a function and a constructor.
        If batch_size is None, creates a non-batch factor.
        If batch_size is an integer, creates a batch factor; in this case, first parameters of function must be
        the batch row index.
        """

        table_class = PyTorchLogTable if log_space else PyTorchTable

        batch = batch_size is not None

        variable_ranges = [range(v.cardinality) for v in variables]
        ranges = variable_ranges if not batch else [range(batch_size)] + variable_ranges

        table_shape = shape(variables) if not batch else (batch_size, *shape(variables))

        table = table_class.from_function(table_shape, ranges, function, batch)

        return constructor(variables, table)

    @staticmethod
    def from_function(variables, function, log_space=True, batch_size=None):
        """
        Builds a PyTorchTableFactor from neuralpp.its variables and a function.
        If batch_size is None, creates a non-batch factor.
        If batch_size is an integer, creates a batch factor; in this case, first parameters of function must be
        the batch row index.
        """
        return PyTorchTableFactor.from_function_and_constructor(
            variables,
            function,
            PyTorchTableFactor,
            log_space=log_space,
            batch_size=batch_size,
        )

    @staticmethod
    def from_predicate(variables, predicate, log_space=True, batch_size=None):
        return PyTorchTableFactor.from_function(
            variables, lambda *args: float(predicate(*args)), log_space, batch_size
        )



import math
import os
from itertools import tee
from typing import List

import torch


def split(xs, predicate):
    result = {True: [], False: []}
    for x in xs:
        result[predicate(x)].append(x)
    return result[True], result[False]


def join(xs, sep=", "):
    return sep.join([str(x) for x in xs])


def map(func, iterable):
    return (func(e) for e in iterable)


def map_of_nested_list(f, o):
    if isinstance(o, list):
        return [map_of_nested_list(f, e) for e in o]
    else:
        try:
            return f(o)
        except Exception as e:
            print(f"Error in map_of_nested_list for argument {o} and function {f}: {e}")
            raise e


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def get_or_put(dictionary, key, default):
    if key in dictionary:
        return dictionary[key]
    else:
        dictionary[key] = default
        return default


def vararg_or_array(arguments_list):
    if len(arguments_list) == 1 and type(arguments_list[0]) in {list, tuple}:
        return arguments_list[0]
    else:
        return arguments_list


def find(iterable, predicate):
    return next((x for x in iterable if predicate(x)), None)


def union_of_dicts(dict1, dict2):
    return {**dict1, **dict2}


def rounded_list(tensor, n_digits=3):
    return map_of_nested_list(lambda x: round(x, n_digits), tensor.tolist())


def merge_dicts(*dicts):
    return {x: d[x] for d in dicts for x in d}


def repeat(n, f):
    """Returns a list with the results of invoking f() n times"""
    return [f() for i in range(n)]


def union(sets):
    result = {}
    for s in sets:
        result.update(s)
    return result


def close(v1, v2, tolerance):
    return abs(v2 - v1) <= tolerance


def mean(sequence):
    total = 0
    length = 0
    for value in sequence:
        total += value
        length += 1
    return total / length


def select_indices_fraction(n, fraction):
    assert 0 < fraction <= 1.0, "select_indices_fraction's fraction must be in (0, 1]"
    rate = 1.0 / fraction
    result = []
    i = 0.0
    while i < n:
        result.append(round(i))
        i += rate
    return result


def all_dims_but_first(tensor):
    num_dims = len(tensor.shape)
    return list(range(1, num_dims))


def is_iterable(object):
    try:
        iter(object)
        return True
    except Exception:
        return False


def is_empty_iterable(object):
    if is_iterable(object):
        try:
            next(iter(object))
            return False
        except StopIteration:
            return True
    else:
        return False


def map_iterable_or_value(function, iterable_of_value):
    if is_iterable(iterable_of_value):
        return [function(e) for e in iterable_of_value]
    else:
        return function(iterable_of_value)


def array_shape(array):
    """Computes the shape of tensor that would be created from given array"""
    return torch.tensor(array).shape


def tile_tensor(tensor, n, dim):
    """Replicate tensor n times along dim, without allocating memory (using torch.expand)"""

    assert 0 <= dim < len(tensor.shape), "dim must be a dimension of the given tensor"

    # this function works by unsqueezing tensor at dim (which will have size 1 and can therefore be expanded),
    # expanding dim by n, and reshaping it to the final shape, which is the original shape multiplied by n at dim.
    number_of_dimensions_after_unsqueezing = len(tensor.shape) + 1
    expand_at_dim_arg = [-1] * number_of_dimensions_after_unsqueezing
    expand_at_dim_arg[dim] = n
    final_shape = list(tensor.shape)
    final_shape[dim] *= n
    return tensor.unsqueeze(dim).expand(*expand_at_dim_arg).reshape(*final_shape)


def check_that_exception_is_thrown(thunk, exception_type):
    if isinstance(exception_type, BaseException):
        raise Exception(
            f"check_that_exception_is_thrown received an exception instance rather than an exception type: "
            f"{exception_type}"
        )
    try:
        thunk()
        raise AssertionError(f"Should have thrown {exception_type}")
    except exception_type:
        pass
    except Exception as e:
        raise AssertionError(
            f"Should have thrown {exception_type} but instead threw {e}"
        )


def matrix_from_function(dims: List, function, prefix_index=None):
    if prefix_index is None:
        prefix_index = []
    if len(dims) == 0:
        return function(*prefix_index)
    else:
        first_dim = dims.pop(0)
        result = [
            matrix_from_function(dims, function, prefix_index=prefix_index + [index])
            for index in first_dim
        ]
        dims.insert(0, first_dim)
        return result


def has_len(obj):
    try:
        len(obj)
        return True
    except TypeError:
        return False


def generalized_len(obj):
    if has_len(obj):
        return len(obj)
    else:
        return 1


def set_default_tensor_type_and_return_device(try_cuda, print=print):
    use_cuda = try_cuda and torch.cuda.is_available()
    if use_cuda:
        torch.set_default_tensor_type(torch.cuda.FloatTensor)
        device = torch.device("cuda")
    else:
        torch.set_default_tensor_type(torch.FloatTensor)
        device = torch.device("cpu")
    print(f"Using {device} device")
    return device


def run_noisy_test(
    noisy_test,
    prob_spurious_failure=0.1,
    target_prob_of_unfair_rejection=0.01,
    print=print,
):
    """
    A utility for testing routines that may fail, even if correct, with a small probability (a spurious failure).
    Assumes that a success only happens if the test is indeed passing
    (therefore a single successful run is enough to conclude a passing test).
    It then proceeds to running the noisy test n times, and fails only if all n runs of noisy test fail.
    n is determined based on a wished target probability of an overall unfair rejection
    (that is, that n runs of the noisy test fail even though it is actually correct).
    """
    # Given n successive failures, probability of correct test is = prob_spurious_failure^n.
    # If we decide for overall failure,
    # the probability of being wrong (unfair rejection) is the probability of being correct.
    # prob of testing being correct = target_prob_of_unfair_rejection = p_spurious_failure^n
    # thus n = log_p target_prob_of_unfair_rejection.
    # If n is not integral, we take the ceiling.
    n = math.ceil(math.log(target_prob_of_unfair_rejection, prob_spurious_failure))
    try_noisy_test_up_to_n_times(noisy_test, n, print)


def try_noisy_test_up_to_n_times(noisy_test, n=3, print=print):
    """
    A utility for testing routines that may fail, even if correct, with a small probability (a spurious failure).
    It attempts to get a successful run up to 3 times, and fails in case that does not happen.
    """
    failure_value = None
    for i in range(n):
        try:
            failure_value = noisy_test()
            if failure_value is True:
                failure_value = None
        except AssertionError as e:
            failure_value = e

        if failure_value is None:
            return True

        if print is not None:
            print("Test failed: " + str(failure_value))
            if i < n - 1:
                print("That may have been spurious, so will try again.")

    if isinstance(failure_value, Exception):
        raise failure_value
    else:
        return failure_value


def assert_equal_up_to_relative_tolerance(actual, expected, tol, message=None):
    """
    Compares two numeric values up to a relative tolerance (a non-negative number).
    Succeeds if expected/(1 + tol) <= actual <= expected*(1 + tol)
    """
    try:
        if tol < 0:
            raise ValueError(
                "Tolerance must be a non-negative number but was " + str(tol)
            )
        if not (expected / (1 + tol) <= actual <= expected * (1 + tol)):
            if message is None:
                message = (
                    f"{actual} not equal to {expected} within relative tolerance {tol}"
                )
            raise AssertionError(message)
    except Exception as e:
        if isinstance(e, AssertionError):
            raise e
        else:
            raise ValueError(
                f"equal_to_up_to_relative_tolerance applies to numeric values only, "
                "and tol must be a non-negative number, but got actual = {actual}, expected = {expected}, and tol = {tol}"
            )


def check_path_like(path, caller=None):
    try:
        os.fspath(path)
    except TypeError:
        raise TypeError(
            ("E" if caller is None else f"{caller} e")
            + f"xpected path-like object but got {path}"
        )


def set_seed(seed=None, print=print):
    """
    Set PyTorch seed, using a pseudo-random one by default.
    Prints the seed at the end using given print function.
    """
    if seed is None:
        seed = torch.seed()
    else:
        torch.manual_seed(seed)
    print(f"Seed: {seed}")


def go_up_until_we_have_subdirectory(subdir):
    initial_directory = os.getcwd()
    while not os.path.isdir(subdir):
        if os.getcwd() == "/":
            os.chdir(initial_directory)
            raise FileNotFoundError(f"Cannot find ancestor of '{initial_directory}' containing subdirectory '{subdir}'")
        os.chdir("..")


def value_tensor(value):
    if isinstance(value, torch.Tensor):
        return value
    else:
        return torch.tensor([value], dtype=torch.float)


def expand_into_batch(tensor, batch_size):
    """
    Returns a batch (of the given size) of replicas of given tensor.
    """
    tensor_with_newly_added_batch_dimension = tensor.unsqueeze(dim=0)
    number_of_non_batch_dimensions = tensor.dim()
    expansion_of_first_dimension = [batch_size]
    expansion_of_remaining_dimensions = [-1] * number_of_non_batch_dimensions
    expansions_by_dimensions = expansion_of_first_dimension + expansion_of_remaining_dimensions
    return tensor_with_newly_added_batch_dimension.expand(expansions_by_dimensions)


def cartesian_prod_2d(tensors):
    """
    Returns a tensor of tensors CP,
    where CP_i,j is the j-th element of the i-th tuple
    in the cartesian product of given tensors.
    This differs from torch.cartensian_prod in the case
    'tensors' is a single tensor, because in that case
    torch.cartesian_prod returns the tensor itself,
    while this function returns a tensor of tensor,
    which is consistent with the general case of multiple input tensors.
    """
    cartesian_product = torch.cartesian_prod(*tensors)
    if len(tensors) == 1:
        cartesian_product = cartesian_product.unsqueeze(
            1
        )  # to make sure cartesian_product is always 2D
    return cartesian_product


def cartesian_product_of_two_tensors(tensor1, tensor2):
    """
    Returns a tensor [ cat(tensor11,tensor21), cat(tensor11, tensor22), ..., cat(tensor1n,tensor2m)]
    where tensor1 is [tensor11, tensor12, ..., tensor1n]
    and tensor2 is [tensor21, tensor22, ..., tensor2m].
    The result is placed in the same device as tensor1.
    """
    expanded_tensor1 = tensor1.repeat_interleave(len(tensor2), dim=0)
    expanded_tensor2 = tensor2.repeat(len(tensor1), 1)
    expanded_tensor2 = expanded_tensor2.to(expanded_tensor1.device).detach()
    cartesian_product = torch.cat((expanded_tensor1, expanded_tensor2), dim=1)
    return cartesian_product


def dict_slice(dict, keys):
    return {k: v for k, v in dict.items() if k in keys}


class RepeatFirstDimensionException(BaseException):
    def __init__(self):
        super(RepeatFirstDimensionException, self).__init__(
            "repeat_first_dimension methods require tensor with at least one dimension"
        )


def _check_repeat_first_dimension_conditions(tensor):
    if not (isinstance(tensor, torch.Tensor) and tensor.dim() != 0):
        raise RepeatFirstDimensionException()


def repeat_first_dimension_with_expand(tensor, n):
    _check_repeat_first_dimension_conditions(tensor)
    original_first_dimension_length = tensor.shape[0]
    final_first_dimension_length = n * original_first_dimension_length
    final_shape = (final_first_dimension_length,) + tensor.shape[1:]
    one_d = tensor.reshape(1, tensor.numel(),)
    expanded = one_d.expand(n, *((-1,)*(one_d.dim() - 1)))
    result = expanded.reshape(final_shape)
    return result


def repeat_interleave_first_dimension(tensor, n):
    _check_repeat_first_dimension_conditions(tensor)
    return torch.repeat_interleave(tensor, n, dim=0)


def flatten_one_level(iterable, is_nested, get_nested):
    """
    Returns [get_nested(e) if is_nested(e) else e for e in iterable]
    """
    return [get_nested(e) if is_nested(e) else e for e in iterable]


def isinstance_predicate(type):
    return lambda o: isinstance(o, type)



from neuralpp.util.group import Group


class BruteForce:
    @staticmethod
    def run(query, factors):
        total_product = Group.product(factors)
        other_variables = {v for f in factors for v in f.variables if v != query}
        marginal = (total_product ^ other_variables).normalize()
        return marginal



import matplotlib.pyplot as plt
import numpy as np

# 100 linearly spaced numbers
import torch


x = np.linspace(-5, 5, 100)

# the function, which is y = x^2 here
y = torch.sigmoid(torch.tensor(x))

# setting the axes at the centre
fig = plt.figure()
ax = fig.add_subplot(1, 1, 1)
ax.spines["left"].set_position("center")
ax.spines["bottom"].set_position("zero")
ax.spines["right"].set_color("none")
ax.spines["top"].set_color("none")
ax.xaxis.set_ticks_position("bottom")
ax.yaxis.set_ticks_position("left")

# plot the function
plt.plot(x, y, "r")

# show the plot
plt.show()



import functools

import torch

from neuralpp.inference.graphical_model.representation.factor.atomic_factor import AtomicFactor
from neuralpp.inference.graphical_model.representation.factor.product_factor import ProductFactor
from neuralpp.util import util
from neuralpp.util.util import join


class ContinuousFactor(AtomicFactor):
    """An abstract specialization of AtomicFactor with at least one continuous variable"""

    def __init__(self, variables, conditioning_dict=None):
        super().__init__(variables)
        self.conditioning_dict = conditioning_dict if conditioning_dict else {}

    def total_conditioning_dict(self, assignment_dict):
        return util.union_of_dicts(
            assignment_dict, self.conditioning_dict
        )

    def assignments(self):
        self._invalid_method("assignments")

    @property
    @functools.lru_cache(1)
    def table_factor(self):
        self._invalid_method("table_factor")

    @staticmethod
    def _invalid_method(method_name):
        error = NotImplementedError(f"Factor.{method_name} invalid for factors with continuous variables")
        raise error

    def mul_by_non_identity(self, other):
        return ProductFactor(self, other)

    def sum_out_variable(self, variable):
        self._not_implemented("sum_out_variable")

    def argmax(self):
        self._not_implemented("argmax")

    def normalize(self):
        self._not_implemented("normalize")

    def sample(self):
        self._not_implemented("sample")

    def _not_implemented(self, method_name):
        error = NotImplementedError(f"Factor.{method_name} not yet supported for {type(self)}")
        raise error



import random

from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import (
    GraphicalModelSGDLearner,
)
from neuralpp.inference.graphical_model.representation.factor.fixed.fixed_pytorch_factor import (
    FixedPyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.factor.neural.MLPFactor import (
    MLPFactor,
)
from neuralpp.inference.graphical_model.representation.frame.multi_frame_data_loader import (
    MultiFrameDataLoader,
)
from neuralpp.inference.graphical_model.representation.model.model import (
    cross_entropy_for_dataset,
)
from neuralpp.inference.graphical_model.representation.random.random_dataset import (
    generate_dataset,
)
from neuralpp.inference.graphical_model.representation.random.random_model import (
    generate_model,
)
from neuralpp.util import util


def run_tests(
    number_of_tests,
    number_of_factors,
    number_of_variables,
    cardinality,
    fraction_of_replaced_factors,
    number_of_sets_of_observed_and_query_variables,
    number_of_query_variables,
    number_of_observed_variables,
    number_of_observations_per_random_set_of_observed_and_query_variables,
    datapoints_per_observation,
    loss_decrease_tol=1e-3,
    cross_entropy_ratio_tol=1.2,
    required_percentage=0.75,
):

    test_results = [
        one_test(
            f"Starting test {i}",
            number_of_factors,
            number_of_variables,
            cardinality,
            fraction_of_replaced_factors,
            number_of_sets_of_observed_and_query_variables,
            number_of_query_variables,
            number_of_observed_variables,
            number_of_observations_per_random_set_of_observed_and_query_variables,
            datapoints_per_observation,
            loss_decrease_tol,
            cross_entropy_ratio_tol,
        )
        for i in range(number_of_tests)
    ]

    number_of_passing_tests = sum(test_results)

    print(f"Successful tests: {number_of_passing_tests} out of {number_of_tests}")

    assert number_of_passing_tests > required_percentage * number_of_tests, (
        f"Only {number_of_passing_tests} passed among {number_of_tests} tests. "
        f"(required percentage: {required_percentage*100:.2}%). "
        f"This might have been either a fluke or a real problem with the code."
    )


def one_test(
    message,
    number_of_factors,
    number_of_variables,
    cardinality,
    fraction_of_replaced_factors,
    number_of_sets_of_observed_and_query_variables,
    number_of_query_variables,
    number_of_observed_variables,
    number_of_observations_per_random_set_of_observed_and_query_variables,
    datapoints_per_observation,
    loss_decrease_tol=1e-3,
    cross_entropy_ratio_tol=1.2,
):

    assert (
        number_of_query_variables + number_of_observed_variables <= number_of_variables
    ), f"number of query variables + number of observed variables must be less than number of variables, but got {number_of_query_variables}, {number_of_observed_variables}, {number_of_variables} respectively"

    print(message)

    ground_truth_model, learned_model = None, None

    while learned_model is None:
        ground_truth_model, learned_model = make_models(
            number_of_factors,
            number_of_variables,
            cardinality,
            fraction_of_replaced_factors,
        )

        debug_model("Ground truth model:", ground_truth_model)
        debug_model("Initial state of model to be learned:", learned_model)

        number_of_replaced_factors = sum(
            [l_f is not g_f for l_f, g_f in zip(learned_model, ground_truth_model)]
        )

        print("Number of ground truth factors:", len(ground_truth_model))
        print("Number of replaced factors:", number_of_replaced_factors)

        if number_of_replaced_factors == 0:
            learned_model = None

    dataset = generate_dataset(
        ground_truth_model,
        number_of_sets_of_observed_and_query_variables,
        number_of_query_variables,
        number_of_observed_variables,
        number_of_observations_per_random_set_of_observed_and_query_variables,
        datapoints_per_observation,
    )

    print(
        f"Generated random dataset from neuralpp.ground truth model of size {len(dataset)}"
    )

    ground_truth_cross_entropy = cross_entropy_for_dataset(
        dataset, ground_truth_model
    ).item()

    print(f"Ground truth model cross entropy: {ground_truth_cross_entropy :.3f}")

    data_loader = MultiFrameDataLoader(dataset, batch_size=10)
    cross_entropy = GraphicalModelSGDLearner(
        learned_model, data_loader, loss_decrease_tol=loss_decrease_tol
    ).learn()

    print(f"Ground truth model cross entropy: {ground_truth_cross_entropy:.3f}")
    print(f"Learned model cross entropy     : {cross_entropy:.3f}")

    debug_model("Learned model after learning is completed:", learned_model)
    # print(join([(type(f).__name__, f) for f in learned_model], "\n"))

    expected_cross_entropy_upper_bound = (
        ground_truth_cross_entropy * cross_entropy_ratio_tol
    )
    if cross_entropy <= expected_cross_entropy_upper_bound:
        print("Test successful!")
        return True
    else:
        print(
            f"Learning performance was poorer than expected. "
            f"Expected upper bound: {expected_cross_entropy_upper_bound:0.3f}, "
            f"actual: {cross_entropy:.3f}"
        )
        return False


def debug_model(message, model):
    pass
    # print(message)
    # for f in model:
    #     print(type(f).__name__)
    #     print(f)
    #     print("Parameters:", list(f.pytorch_parameters()))


def make_models(
    number_of_factors, number_of_variables, cardinality, fraction_of_replaced_factors
):
    ground_truth_model = generate_model(
        number_of_factors, number_of_variables, cardinality, FixedPyTorchTableFactor
    )
    selected_for_replacement = set(
        util.select_indices_fraction(number_of_factors, fraction_of_replaced_factors)
    )
    learned_model = [
        replace(ground_truth_model[i])
        if i in selected_for_replacement
        else ground_truth_model[i]
        for i in range(number_of_factors)
    ]
    return ground_truth_model, learned_model


def replace(factor):
    if len(factor.variables) > 1:
        return make_neural_factor(factor).randomized_copy()
    else:
        return factor


def make_neural_factor(factor):
    assert (
        len(factor.variables) > 0
    ), f"To replace a factor by a neural factor, the original must have at least one variable, but got a factor without none: {factor}"
    output_variable = random.choice(factor.variables)
    input_variables = [v for v in factor.variables if v != output_variable]
    number_of_hidden_units = len(input_variables)
    neural_factor = MLPFactor.make(
        input_variables, number_of_hidden_units, output_variable
    )
    return neural_factor


def test_random_joint_learning():
    run_tests(
        number_of_tests=2,
        number_of_factors=8,
        number_of_variables=8,
        cardinality=3,
        fraction_of_replaced_factors=0.5,
        number_of_sets_of_observed_and_query_variables=10,
        number_of_query_variables=1,
        number_of_observed_variables=2,
        number_of_observations_per_random_set_of_observed_and_query_variables=10,
        datapoints_per_observation=1,
        loss_decrease_tol=1e-2,
        cross_entropy_ratio_tol=1.2,
    )



import random
import sys

import torch
from neuralpp.inference.graphical_model.brute_force import BruteForce
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.random.random_model import (
    generate_model,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable_elimination import VariableElimination


print("sys.path:", sys.path)


def run_test_ve_against_brute_force(query, model):
    for (i, factor) in enumerate(model):
        print(f"Factor {i}", model[i])
    ve_result = VariableElimination().run(query, model)
    print(f"\nVE marginal on {query}:", ve_result)
    brute_result = BruteForce().run(query, model)
    print(f"\nBrute force marginal on {query}:", brute_result)
    assert ve_result == brute_result


def test_variable_elimination():

    for log_space in {False, True}:

        x = IntegerVariable("x", 3)
        y = IntegerVariable("y", 2)
        z = IntegerVariable("z", 2)

        # noinspection PyShadowingNames
        model = [
            PyTorchTableFactor(
                [], torch.ones(()), log_space=log_space
            ),  # just to test the edge case
            PyTorchTableFactor([x], [0.4, 0.1, 0.5], log_space=log_space),
            PyTorchTableFactor.from_function(
                [y, x],
                lambda y, x: 0.5 if x == 2 else 1.0 if y == x else 0.0,
                log_space=log_space,
            ),
            PyTorchTableFactor.from_function(
                [z, y], lambda z, y: 1.0 if z == y else 0.0, log_space=log_space
            ),
        ]

        query = z

        run_test_ve_against_brute_force(query, model)

    for i in range(10):
        model = generate_model(
            number_of_factors=15, number_of_variables=8, cardinality=2
        )
        query = random.choice([v for f in model for v in f.variables])
        print()
        run_test_ve_against_brute_force(query, model)



from types import SimpleNamespace

import torch

from neuralpp.experiments.mnist_semi_supervised.mnist_semi_supervised import default_parameters, \
    solve_learning_problem_from_parameters, make_digits_and_all_true_constraints_values_batches_generator
from neuralpp.inference.graphical_model.learn.learning_problem_solver import solve_learning_problem, LearningProblem
from neuralpp.inference.graphical_model.learn.uniform_training import UniformTraining
from neuralpp.inference.graphical_model.representation.factor.fixed.fixed_pytorch_factor import (
    FixedPyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.table.pytorch_log_table import (
    PyTorchLogTable,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable.tensor_variable import TensorVariable
from neuralpp.inference.neural_net.ConvNet import ConvNet
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.inference.neural_net.from_log_to_probabilities_adapter import (
    FromLogToProbabilitiesAdapter,
)
from neuralpp.util.data_loader_from_random_data_point_thunk import (
    data_loader_from_batch_generator,
)
from neuralpp.util.generic_sgd_learner import default_after_epoch
from neuralpp.util.mnist_util import read_mnist, show_images_and_labels
from neuralpp.util.util import join, set_seed

# Trains a digit recognizer with the following factor graph:
#
#                                 Constraint0                                   Constraint1
#                                      |                                             |
#                           +---------------------+                       +---------------------+
#                           |   Constraint0 <=>   |                       |   Constraint1 <=>   |
#             Digit0  ------| Digit1 = Digit0 + 1 |------- Digit1 --------| Digit2 = Digit1 + 1 |------- ...
#               |           +---------------------+          |            +---------------------+
#               |                                            |
#      +------------------+                         +------------------+
#      | Digit recognizer |                         | Digit recognizer |
#      +------------------+                         +------------------+
#               |                                            |
#               |                                            |
#             Image0                                       Image1
#
# that is to say, (3n - 1) variables
# Constraint_0, ..., Constraint_(n-2), Digit0, ..., Digit_(n-1), Image0, ... Image_(n-1),
# where Digit_i is constrained to be Digit_(i-1) + 1 iff Constraint_(i-1) is true,
# and Digit_i is the recognition of Image_i.
#
# The original, harder and more interesting case is for n = 2.
#
# We anticipate that presenting pairs images of consecutive digits to the model,
# querying Constraint and minimizing its epoch_average_loss in comparison to the expected value "true"
# will train the recognizer (a shared neural net applied to both images)
# to recognize MNIST digits.
#
# The idea is that images for 0 will only appear in Image0 (because it has no antecessor),
# and images for 9 only in Image_(n-1) (because it has no successor),
# and this will be a strong enough signal for the recognizer to learn those images.
# Once that happens, those images work as the signal for 1 and 8, which work as signals for 2 and 7
# and so on, until all digits are learned.
#
# There are multiple ways of running this experiment. See mnist_semi_supervised.py for extra options.


def indices_of_digit_arguments_of_constraint(constraint_index):
    i = constraint_index
    return i, i + 1


def constraint_function(constraint_index, di, di_plus_one):
    # all constraints use the same predicate, so constraint_index is irrelevant.
    return di_plus_one == di + 1


def generate_chain_of_successive_digits_batch(number_of_digits, chain_length, batch_size):
    """
    Must generate an example guaranteed to satisfy all constraints
    """
    digit_values = []
    last_digit = number_of_digits - 1
    number_of_transitions_in_the_chain = chain_length - 1
    max_value_of_first_digit = last_digit - number_of_transitions_in_the_chain
    exclusive_upper_bound_for_first_digit = max_value_of_first_digit + 1
    digit_values.append(
        torch.randint(exclusive_upper_bound_for_first_digit, (batch_size,))
    )
    for i in range(1, chain_length):
        digit_values.append(digit_values[i - 1] + 1)
    return digit_values


chain_of_successive_digits_and_all_true_constraints_batch_generator = \
    make_digits_and_all_true_constraints_values_batches_generator(
        generate_chain_of_successive_digits_batch
    )

parameters = default_parameters()
parameters.chain_length = 4
parameters.number_of_constraints = parameters.chain_length - 1
parameters.number_of_constraint_values = 2  # boolean constraints
parameters.indices_of_digit_arguments_of_constraint = indices_of_digit_arguments_of_constraint
parameters.constraint_function = constraint_function

parameters.custom_digits_and_constraints_values_batches_generator = None
#parameters.custom_digits_and_constraints_values_batches_generator = chain_of_successive_digits_and_all_true_constraints_batch_generator

solve_learning_problem_from_parameters(parameters)



import itertools
import math

import torch
from neuralpp.inference.graphical_model.representation.representation import (
    contains_multivalue_coordinate,
    is_multivalue_coordinate,
)
from neuralpp.inference.graphical_model.representation.table.table import Table
from neuralpp.inference.graphical_model.representation.table.table_util import (
    insert_shape,
    n_unsqueeze,
)
from neuralpp.util import util
from neuralpp.util.batch_argmax import batch_argmax
from neuralpp.util.first import first
from neuralpp.util.tensor_mixed_radix import TensorMixedRadix
from neuralpp.util.util import all_dims_but_first, is_iterable, map_of_nested_list
from torch.distributions import Categorical


class PyTorchTable(Table):
    def __init__(self, raw_entries, batch=False):
        super().__init__()
        self.raw_tensor = (
            raw_entries
            if isinstance(raw_entries, torch.Tensor)
            else torch.tensor(raw_entries, requires_grad=True)
        )
        self.batch = batch
        if self.batch:
            self.non_batch_shape = self.raw_tensor.shape[1:]
        else:
            self.non_batch_shape = self.raw_tensor.shape

        self._non_batch_radices = None

    def new_table_from_raw_entries(self, raw_entries, batch=False):
        return type(self)(raw_entries, batch)

    def shape(self):
        return self.raw_tensor.shape

    def number_of_batch_rows(self):
        """Returns the number of batch rows or None if table is not a batch"""
        if self.batch:
            return self.raw_tensor.shape[0]
        else:
            return None

    @staticmethod
    def from_array(array_of_potentials, batch=False):
        return PyTorchTable(array_of_potentials, batch)

    @staticmethod
    def from_function(
        shape, function_arguments_iterables, function_of_potentials, batch=False
    ):

        array_of_potentials = []
        for function_arguments_tuple in itertools.product(
            *function_arguments_iterables
        ):
            potential = function_of_potentials(*function_arguments_tuple)
            array_of_potentials.append(float(potential))

        # array_of_potentials = [float(function_of_potentials(*function_arguments_tuple))
        #                        for function_arguments_tuple in itertools.product(*function_arguments_iterables)]

        one_dim_tensor = torch.tensor(array_of_potentials)
        right_shape_tensor_of_potentials = one_dim_tensor.reshape(shape)
        return PyTorchTable(right_shape_tensor_of_potentials, batch)

    # Methods depending on structure only (not values)
    # This distinction is useful if a sub-class represents values in some other system
    # (such as in log-space), because then we know these methods need not be overridden.

    def __len__(self):
        return self.raw_tensor.numel()

    def assignments(self):
        return itertools.product(*[range(dim) for dim in self.non_batch_shape])

    def expand(self, shape_to_be_inserted, dim):
        """
        Inserts new dimensions with a given shape at specific position dim
        (after disregarding possible batch dimension)
        """
        effective_dim = dim + 1 if self.batch else dim
        unsqueezed_tensor = n_unsqueeze(
            self.raw_tensor, len(shape_to_be_inserted), effective_dim
        )
        final_shape = insert_shape(
            self.raw_tensor.shape, shape_to_be_inserted, effective_dim
        )
        new_tensor = unsqueezed_tensor.expand(final_shape)
        return self.new_table_from_raw_entries(new_tensor, self.batch)

    def permute(self, permutation):
        """
        Applies permutation (of non-batch dimensions).
        """
        effective_permutation = (
            [0] + [p + 1 for p in permutation] if self.batch else permutation
        )
        permuted_raw_tensor = self.raw_tensor.permute(effective_permutation)
        return self.new_table_from_raw_entries(permuted_raw_tensor, self.batch)

    def pytorch_parameters(self):
        return [self.raw_tensor]

    def get_raw_tensor_slice(self, non_batch_slice_coordinates):
        # see the documentation for Table.__getitem__ to better understand this implementation
        if is_iterable(non_batch_slice_coordinates):
            tuple_of_non_batch_slice_coordinates = tuple(non_batch_slice_coordinates)
        else:
            tuple_of_non_batch_slice_coordinates = (non_batch_slice_coordinates,)

        if self.batch:
            # pick the value for the assignment in each batch row
            batch_rows_coordinate = self.get_batch_rows_coordinate(
                tuple_of_non_batch_slice_coordinates
            )
            all_coordinates = (
                batch_rows_coordinate,
                *tuple_of_non_batch_slice_coordinates,
            )
        else:
            all_coordinates = tuple_of_non_batch_slice_coordinates

        self.check_all_multivalue_coordinates_are_1d_and_have_the_same_size(all_coordinates)

        return self.raw_tensor[all_coordinates]

    def get_batch_rows_coordinate(self, tuple_of_non_batch_slice_coordinates):
        # TODO: Using range is much slower than using slice(None)
        # however slice(None) does not produce the same result if there are batch coordinates.
        # For example, assume a 2D n x n tensor A.
        # A[range(n), range(n)] is its diagonal,
        # A[slice(None), range(n)] is equal to A.
        # However, if there are no batch coordinates then we can use slice and obtain the same result,
        # hence the case analysis used here.
        if any(is_multivalue_coordinate(c) for c in tuple_of_non_batch_slice_coordinates):
            batch_rows_coordinate = range(self.number_of_batch_rows())
        else:
            batch_rows_coordinate = slice(None)
        return batch_rows_coordinate

    def check_all_multivalue_coordinates_are_1d_and_have_the_same_size(
        self, all_coordinates
    ):
        multivalue_coordinates = [c for c in all_coordinates if is_multivalue_coordinate(c)]

        invalid_multivalue_coordinate = first(
            multivalue_coordinates, lambda bc: len(bc) != 0 and is_iterable(bc[0])
        )
        if invalid_multivalue_coordinate is not None:
            raise BatchCoordinateFirstElementIsIterable(invalid_multivalue_coordinate)

        set_of_len_of_multivalue_coordinates = {
            len(multivalue_coordinate) for multivalue_coordinate in multivalue_coordinates
        }

        if len(set_of_len_of_multivalue_coordinates) > 1:
            raise BatchCoordinatesDoNotAgreeException()

    def slice(self, non_batch_slice_coordinates):
        raw_tensor_slice = self.get_raw_tensor_slice(
            non_batch_slice_coordinates
        )  # already covers batch cases
        new_table_is_batch = self.batch or contains_multivalue_coordinate(
            non_batch_slice_coordinates
        )
        return self.new_table_from_raw_entries(raw_tensor_slice, new_table_is_batch)

    # Methods depending on structure and value representation choice (here, normal space rather than log)

    # This distinction is useful if a sub-class represents values in some other system
    # (such as in log-space), because then we know these methods do need to be overridden.

    def __getitem__(self, non_batch_slice_coordinates):
        """Does what super method does, with the additional fact that the returned value is a torch.Tensor"""
        return self.get_raw_tensor_slice(non_batch_slice_coordinates)

    def __mul__(self, other):
        """
        Multiply by another table, taking care of batch concerns.
        Delegates actually multiplication of raw tensors
        so that sub-classes can reuse this method and override that method only.
        """
        assert (
            self.non_batch_shape == other.non_batch_shape
        ), f"Multiplied {type(self)} instances must have the same non-batch shapes, but have {self.non_batch_shape} and {other.non_batch_shape}"
        assert (
            not (self.batch and other.batch)
            or self.number_of_batch_rows() == other.number_of_batch_rows()
        ), f"If both tables are batches then they must have the same number of rows, but they have {self.number_of_batch_rows()} and {other.number_of_batch_rows()} rows respectively"

        result_is_batch = self.batch or other.batch

        if type(self) == type(other):
            # if 'other' is the same type as self, more assumptions can be made about raw tensors,
            # allowing greater efficiency -- the assumptions are encapsulated
            # in the implementation of method raw_tensor_of_product_of_potentials_of_raw_tensors
            raw_tensor = self.raw_tensor_of_product_of_potentials_of_raw_tensors(
                self.raw_tensor, other.raw_tensor
            )
            return self.new_table_from_raw_entries(raw_tensor, result_is_batch)
        else:
            # otherwise, pay the penalty of possibly converting back and forth from neuralpp.non-potential raw tensors.
            # Note: operating on potentials is not less efficiency for class PyTorchTable,
            # but is be for sub-classes using different representations, such as PyTorchLogTable
            array_of_potentials = self.potentials_tensor() * other.potentials_tensor()
            return self.from_array(array_of_potentials, result_is_batch)

    @staticmethod
    def raw_tensor_of_product_of_potentials_of_raw_tensors(raw_tensor_1, raw_tensor_2):
        """
        Returns the raw_tensor representing the product of potentials corresponding to two raw tensors
        (raw tensors must come from neuralpp.the same class).
        """
        return raw_tensor_1 * raw_tensor_2

    def sum_out(self, dim):
        if util.is_empty_iterable(dim) and self.batch:
            # PyTorch treats an empty sequence of dims to mean "all dims"
            # If dim is an empty list, so will be effective_dim.
            # If we have a batch, using torch.sum on the raw tensor with empty dim will sum everything, including
            # the batch rows, when we actually want the sum to operate over all non-batch dims only but preserve
            # the batch dimension.
            # So we must replace dim by [0,...,len(non_batch_shape) - 1]
            # (later this will be transformed into effective_dim that takes the batch dimension into account)
            dim = list(range(len(self.non_batch_shape)))

        if self.batch:

            def effective_dimension_value(d):
                return d + 1

        else:

            def effective_dimension_value(d):
                return d

        effective_dim = util.map_iterable_or_value(effective_dimension_value, dim)

        potential_tensor = torch.sum(self.raw_tensor, effective_dim)
        return self.from_array(potential_tensor, self.batch)

    def sum(self):
        if self.batch:
            return torch.sum(self.raw_tensor, dim=all_dims_but_first(self.raw_tensor))
        else:
            return torch.sum(self.raw_tensor)

    def argmax(self):
        indices = batch_argmax(self.raw_tensor, batch_dim=1 if self.batch else 0)
        return indices

    def normalize(self):
        if self.batch:
            # we want to divide each row by its sum, which is provided as a one-dimensional tensor by self.sum()
            # To get the proper broadcasting to work, we must reshape it with trailing dimensions of size 1
            # for the non-batch dimensions.
            trailing_size_one_dimensions = [1] * len(self.non_batch_shape)
            n_rows = self.raw_tensor.shape[0]
            partitions = self.sum().reshape(n_rows, *trailing_size_one_dimensions)
            normalized_potential_tensor = self.raw_tensor / partitions
            return self.from_array(normalized_potential_tensor, self.batch)
        else:
            normalized_potential_tensor = self.raw_tensor / self.sum()
            return self.from_array(normalized_potential_tensor)

    def sample(self):
        batch_size = self.number_of_batch_rows() if self.batch else 1
        non_batch_size = math.prod(self.non_batch_shape)
        batch_of_potentials_in_rows = self.potentials_tensor().reshape(
            batch_size, non_batch_size
        )
        sampled_non_batch_assignment_indices = Categorical(
            batch_of_potentials_in_rows
        ).sample()
        assignments = self.from_tensor_of_non_batch_assignment_indices_to_tensor_of_non_batch_assignments(
            sampled_non_batch_assignment_indices
        )
        return assignments

    def from_tensor_of_non_batch_assignment_indices_to_tensor_of_non_batch_assignments(
        self, non_batch_assignment_indices
    ):
        """Takes a 1D tensor of non-batch assignment indices and returns the corresponding assignments"""
        return self.non_batch_radices.representation(non_batch_assignment_indices)

    @property
    def non_batch_radices(self):
        if self._non_batch_radices is None:
            self._non_batch_radices = TensorMixedRadix(self.non_batch_shape)
        return self._non_batch_radices

    def randomize(self):
        self.raw_tensor = torch.rand(self.raw_tensor.shape, requires_grad=True)

    def randomized_copy(self):
        potential_tensor = torch.rand(self.raw_tensor.shape, requires_grad=True)
        return self.from_array(potential_tensor, self.batch)

    def __eq__(self, other):
        """
        Indicates whether two tables are equal, including whether they are batches or not and, if so,
        of the same size. Once that is checked, their potentials are compared with
        torch.allclose with atol=1e-3 and rtol=1e-3
        """
        if isinstance(other, PyTorchTable):
            result = self.has_same_batch_and_shape(other) and torch.allclose(
                self.potentials_tensor(),
                other.potentials_tensor(),
                atol=1e-3,
                rtol=1e-3,
            )
            return result
        else:
            raise Exception(
                "Comparison of PyTorchTable to tables other than PyTorchTable not implemented"
            )

    def has_same_batch_and_shape(self, other):
        return (
            self.number_of_batch_rows() == other.number_of_batch_rows()
            and self.non_batch_shape == other.non_batch_shape
        )

    def __repr__(self):
        """
        Prints a representation for the current table.
        The method uses method 'potentials_tensor' so sub-classes only need to override that to re-use this method.
        """
        return ("batch " if self.batch else "") + str(
            map_of_nested_list(lambda v: round(v, 4), self.potentials_tensor().tolist())
        )

    def potentials_tensor(self):
        return self.raw_tensor


class BatchCoordinateFirstElementIsIterable(Exception):
    """
    Exception throw when indexing the table with a batch coordinate whose first element is an iterable.
    This catches mistakes in which the user sends a batch coordinate not containing scalars.
    Only the first element is checked for efficiency (checking all elements would take much longer and
    be unnecessary because if some elements and iterable and others are not a PyTorch exception
    will be raised anyway).
    """

    def __init__(self, invalid_multivalue_coordinate):
        super(BatchCoordinateFirstElementIsIterable, self).__init__(
            f"Batch coordinate first element is iterable; they should all be integral scalars: "
            f"{invalid_multivalue_coordinate[0]}"
        )


class BatchCoordinatesDoNotAgreeException(Exception):
    """
    Exception throw when indexing the table with batch coordinates (that is, coordinates with more than one value
    for batch processing) are not all of the same type or not agreeing with a batch table's number of row (
    in the case the table is a batch).
    """

    def __init__(self):
        super(BatchCoordinatesDoNotAgreeException, self).__init__(
            f"Batch coordinates do not agree with each other or with number of rows in batch table."
        )



from neuralpp.inference.graphical_model.representation.frame.tuple_data_loader import (
    get_data_loader,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)


class MultiFrameDataLoader:
    def __init__(self, dataset, batch_size=100):
        try:
            len(dataset[0]) == 2 and all(
                all(isinstance(v, DiscreteVariable) for v in dataset[0][i].keys())
                for i in {0, 1}
            )
        except TypeError as e:
            raise Exception(
                "A dataset must be a sequence of tuples of two dicts. The first dict is the observation, "
                "mapping variables to their observed values. The second dict is the query label, mapping query "
                "variables to their queried values"
            ) from e

        self.dataset = dataset
        self.batch_size = batch_size

    def __iter__(self):
        return self.multi_frame_data_loader()

    def multi_frame_data_loader(self):
        # We unusually have two loops over the data: one for pairs (observation_frame, query_assignment_frame)
        # and another for the possibly multiple values associated to random variables in each of those
        # two maps.
        # (For those familiar with R or pandas, it is useful to think of both observation_frame
        # or query_assignment_frame as data frames.)
        # This happens here because we can train the same effective_model from frames defined on different
        # observed and queried variables.
        # The dataset is composed of frames, each defined on the same variables.
        # The usual training is therefore performed for each frame.
        for (observation_frame, query_assignment_frame) in self.dataset:
            data_loader = get_data_loader(
                observation_frame, query_assignment_frame, self.batch_size
            )
            for (observation_batch, query_assignment_batch) in data_loader:
                yield observation_batch, query_assignment_batch



import torch

from neuralpp.inference.graphical_model.representation.factor.continuous.pytorch_distribution_factor import \
    PyTorchDistributionFactor


class NormalFactor(PyTorchDistributionFactor):

    def __init__(self, variables, conditioning_dict=None):
        super().__init__(torch.distributions.Normal, variables, conditioning_dict)



import random

import torch
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.table.table_util import shape
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.util.util import join


def generate_model(
    number_of_factors, number_of_variables, cardinality, factor_class=PyTorchTableFactor
):
    """
    Makes a random model according to parameters, constructing factors using factor_class' from_function method.
    """
    variables = {
        IntegerVariable(f"x{i}", cardinality) for i in range(number_of_variables)
    }
    factors = [
        generate_factor(random_subset(variables), factor_class)
        for i1 in range(number_of_factors)
    ]
    missing_variables = variables - {v for f in factors for v in f.variables}
    uniform_factors_for_missing_variables = [
        factor_class.from_function([m], lambda vm: 1.0) for m in missing_variables
    ]
    return factors + uniform_factors_for_missing_variables


def generate_factor(variables, factor_class):
    return factor_class(variables, torch.rand(shape(variables), requires_grad=True))


def random_subset(variables):
    result = []
    for variable in variables:
        if bool(random.getrandbits(1)):
            result.append(variable)
    return result


if __name__ == "__main__":
    print(
        join(
            generate_model(number_of_factors=4, number_of_variables=5, cardinality=3),
            "\n",
        )
    )



import random

from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import (
    GraphicalModelSGDLearner,
)
from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.test.slow_tests.graphical_model.representation.factor.neural.neural_factor_test_util import (
    check_and_show_conditional_distributions,
)
from neuralpp.util import util
from neuralpp.util.util import try_noisy_test_up_to_n_times


def make_bernoulli_dataset(x, y):
    global dataset

    def bernoulli(p):
        return int(random.random() < p)

    DATASET_SIZE = 1000
    p = [0.3, 0.9]

    x_value = lambda i: 0 if i < DATASET_SIZE / 2 else 1
    y_value = lambda i: bernoulli(p[x_value(i)])
    dataset = [({x: x_value(i)}, {y: y_value(i)}) for i in range(DATASET_SIZE)]
    random.shuffle(dataset)
    return dataset


def mlp_neural_factor_learning():
    x = IntegerVariable("x", 2)
    y = IntegerVariable("y", 2)
    input_variables = [x]
    output_variable = y
    neural_net = MLP(len(input_variables), output_variable.cardinality)
    neural_net = neural_net.randomized_copy()
    neural_factor = NeuralFactor(neural_net, input_variables, output_variable)

    dataset = make_bernoulli_dataset(x, y)
    print()
    print("Dataset first few lines:")
    print(util.join(dataset[:10], "\n"))
    print()
    y_for_x_0 = [row[1][y] for row in dataset if row[0][x] == 0]
    y_for_x_1 = [row[1][y] for row in dataset if row[0][x] == 1]
    # print("y for x = 0:", y_for_x_0)
    # print("y for x = 1:", y_for_x_1)
    y_mean_for_x_0 = sum(y_for_x_0) / len(y_for_x_0)
    y_mean_for_x_1 = sum(y_for_x_1) / len(y_for_x_1)
    print("Dataset mean y for x = 0:", y_mean_for_x_0)
    print("Dataset mean y for x = 1:", y_mean_for_x_1)

    graphical_model = [neural_factor]
    GraphicalModelSGDLearner(graphical_model, dataset, loss_decrease_tol=1e-3).learn()

    def ground_truth(x, y):
        if x == 0:
            return y_mean_for_x_0 if y == 1 else 1 - y_mean_for_x_0
        else:
            return y_mean_for_x_1 if y == 1 else 1 - y_mean_for_x_1

    check_and_show_conditional_distributions(neural_factor, ground_truth)


def test_joint_learning():
    try_noisy_test_up_to_n_times(mlp_neural_factor_learning, n=3)



def first(iterable, condition=lambda x: True, default=None):
    """
    Returns the first item in the `iterable` that
    satisfies the `condition`.

    If the condition is not given, returns the first item of
    the iterable.

    If there is no first element, the default is returned (None if default is not provided).
    """

    try:
        return next(x for x in iterable if condition(x))
    except StopIteration:
        return default



from neuralpp.inference.graphical_model.representation.factor.factor import Factor
from neuralpp.util import util


class AtomicFactor(Factor):
    def __init__(self, variables):
        super().__init__(variables)

    def atomic_factor(self):
        return self

    def sample(self):
        self._not_implemented("sample")

    def single_sample(self):
        """TODO: deprecated; remove"""
        self._not_implemented("single_sample")

    # Convenience methods

    def single_sample_assignment_dict(self):
        """TODO: deprecated; remove"""
        sampled_assignment = self.single_sample()
        return self.from_assignment_to_assignment_dict(sampled_assignment)



import collections
import functools

from neuralpp.util.group import Group


class Factor:
    def __init__(self, variables):
        self.variables = variables

    def __contains__(self, variable):
        return variable in self.variables

    def __call__(self, assignment_dict):
        assignment_values = self.validate_argument_to_call(assignment_dict)
        return self.call_after_validation(assignment_dict, assignment_values)

    def call_after_validation(self, assignment_dict, assignment_values):
        """
        Receives the original assignment_dict passed to __call__ after validation as
        well as a list of values, which is a sub-product of the validation,
        in the same order as their respective variables in this factor.
        Implementations can make use of either one to compute the potential.
        """
        self._not_implemented("call_after_validation")

    def __getitem__(self, assignment_dict):
        return self(assignment_dict)

    def condition(self, assignment_dict):
        """
        Receives an assignment dict from variables to values and returns the result of conditioning the factor
        by that assignment.
        This default implementation checks if the assignment dict is indeed dict-like
        (instance of collections.abc.Mapping).
        If it is, it also checks whether the assignment dict is empty and, if so, the factor is returned unchanged.
        Otherwise, this default implementation delegates to condition_on_non_empty_dict(assignment_dict)
        for class-specific processing.
        """
        if isinstance(assignment_dict, collections.abc.Mapping):
            if assignment_dict:
                return self.condition_on_non_empty_dict(assignment_dict)
            else:
                return self
        else:
            raise Exception(
                f"Factor being conditioned on non-dict-like object {assignment_dict}"
            )

    def condition_on_non_empty_dict(self, assignment_dict):
        self._not_implemented("condition_on_non_empty_dict")

    def randomize(self):
        """Randomized parameters in-place"""
        self._not_implemented("randomized")

    def randomize_copy(self):
        """
        Returns a representation with same structure but randomized parameters
        IMPORTANT: this will create new copies of internal structures, de-coupling parameters
        shared with other factors.
        """
        self._not_implemented("randomized_copy")

    def pytorch_parameters(self):
        self._not_implemented("pytorch_parameters")

    def __mul__(self, other):
        if other is Group.identity:
            return self
        else:
            return self.mul_by_non_identity(other)

    __rmul__ = __mul__

    def mul_by_non_identity(self, other):
        self._not_implemented("mul_by_non_identity")

    def __xor__(self, variable_or_variables):
        if type(variable_or_variables) in [set, list, tuple]:
            return self.sum_out_variables(variable_or_variables)
        else:
            return self.sum_out_variable(variable_or_variables)

    def sum_out_variables(self, variables):
        current = self
        for v in variables:
            current = current ^ v
        return current

    def sum_out_variable(self, variable):
        self._not_implemented("sum_out_variable")

    def argmax(self):
        """
        Returns an assignment dictionary for the largest value in the factor.
        If the factor is a batch, the assignment dict will also have batch values.
        """
        self._not_implemented("argmax")

    def normalize(self):
        self._not_implemented("normalize")

    def atomic_factor(self):
        self._not_implemented("atomic_factor")

    @property
    @functools.lru_cache(1)
    def table_factor(self):
        """Table factor equivalent to self"""
        self._not_implemented("table_factor property")

    def _not_implemented(self, name):
        # creating a variable first prevents compiler from thinking this is an abstract method
        error = NotImplementedError(f"{name} not implemented for {type(self)}")
        raise error

    # Convenience methods

    def from_assignment_to_assignment_dict(self, assignment):
        return dict(zip(self.variables, assignment))

    def validate_argument_to_call(self, assignment_dict):
        try:
            assignment_values = [assignment_dict[v] for v in self.variables]
        except TypeError as e:
            if isinstance(assignment_dict, (list, tuple)):
                raise Exception(
                    f"Factors must be called on an assignment dict from variables to values, not on a {type(assignment_dict).__name__}: {assignment_dict}"
                ) from e
            else:
                raise Exception(
                    f"Type error computing the application of a {type(self).__name__} to an assignment, "
                    f"which must be a dict. The argument was "
                    f"{assignment_dict}"
                ) from e
        except KeyError as e:
            raise Exception(
                f"Indexed access to Factor requires a complete assignment to its variables {self.variables}, but got {assignment_dict} which does not include a variable: {e}"
            )

        return assignment_values



class EveryKTimes:
    """Provides an object that runs a given thunk (0-args function) every k-th time it is invoked"""

    def __init__(self, thunk, k):
        self.thunk = thunk
        self.k = k
        self.counter = 0

    def __call__(self):
        self.counter += 1
        if self.counter == self.k:
            self.counter = 0
            self.thunk()



import torch


def is_multivalue_coordinate(coordinate):
    """
    Indicates whether given coordinate (index in a table or argument to a function) is a batch
    """
    return isinstance(coordinate, (list, range, torch.Tensor))


def contains_multivalue_coordinate(coordinates):
    """
    Indicates whether any of given coordinates is a batch
    """
    return any(is_multivalue_coordinate(c) for c in coordinates)



import itertools

from neuralpp.inference.graphical_model.variable.variable import Variable


class DiscreteVariable(Variable):
    @staticmethod
    def assignments_product(variables):
        if len(variables) != 0:
            return itertools.product(*[v.assignments() for v in variables])
        else:
            return [tuple()]

    @staticmethod
    def assignments_product_dicts(variables):
        def make_dict(values):
            return {var: val for var, val in zip(variables, values)}

        return map(make_dict, DiscreteVariable.assignments_product(variables))

    def __init__(self, name=None, cardinality=None):
        super().__init__()
        self._name = name
        self._cardinality = cardinality

    @property
    def cardinality(self):
        return self._cardinality

    @cardinality.setter
    def cardinality(self, new_value):
        self._cardinality = new_value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_value):
        self._name = new_value

    def assignments(self):
        self._not_implemented("assignments")

    def _not_implemented(self, name):
        error = NotImplementedError(f"{name} not implemented for {type(self)}")
        raise error



import torch
from neuralpp.inference.graphical_model.representation.table.pytorch_table import (
    PyTorchTable,
)
from neuralpp.util.log_util import (
    log_of_nested_list_without_inf_non_differentiable,
    log_without_inf_non_differentiable,
)


class PyTorchLogTable(PyTorchTable):
    def __init__(self, raw_entries, batch=False):
        super().__init__(raw_entries, batch)
        assert not torch.flatten(torch.isnan(self.raw_tensor)).any()

    @staticmethod
    def from_array(array_of_potentials, batch=False):
        if isinstance(array_of_potentials, torch.Tensor):
            tensor = log_without_inf_non_differentiable(array_of_potentials)
        else:
            log_of_potentials = log_of_nested_list_without_inf_non_differentiable(
                array_of_potentials
            )
            tensor = torch.tensor(log_of_potentials, requires_grad=True)
        table = PyTorchLogTable(tensor, batch)
        return table

    @staticmethod
    def from_function(
        shape, function_arguments_iterables, function_of_potentials, batch=False
    ):
        prob_space = PyTorchTable.from_function(
            shape, function_arguments_iterables, function_of_potentials, batch
        )
        log_probability_tensor = log_without_inf_non_differentiable(
            prob_space.raw_tensor
        )
        leaf_tensor = log_probability_tensor.clone().detach().requires_grad_(True)
        return PyTorchLogTable(leaf_tensor, batch)

    # Methods depending on structure and value representation choice (here, log)

    def __getitem__(self, non_batch_slice_coordinates):
        return super().__getitem__(non_batch_slice_coordinates).exp()

    @staticmethod
    def raw_tensor_of_product_of_potentials_of_raw_tensors(raw_tensor_1, raw_tensor_2):
        return raw_tensor_1 + raw_tensor_2  # + since we are at log-space

    def sum_out(self, dim):

        # We resort to PyTorchTable so we can rely on its ability to manage batches
        normal_space_table = PyTorchTable(self.raw_tensor.exp(), self.batch)
        normal_summed_out = normal_space_table.sum_out(dim)

        # We can use log directly here because the potentials are guaranteed not to be exactly 0
        # since they are sums of potentials coming from neuralpp.exponentiations.
        summed_out_log_raw_tensor = normal_summed_out.raw_tensor.log()

        return PyTorchLogTable(summed_out_log_raw_tensor, self.batch)

    def sum(self):
        normal_space_table = PyTorchTable(self.raw_tensor.exp(), self.batch)
        return normal_space_table.sum()

    def normalize(self):
        normal_space_table = PyTorchTable(self.raw_tensor.exp(), self.batch)
        normalized_in_normal_space = normal_space_table.normalize()
        normalized_in_normal_space_raw_tensor = normalized_in_normal_space.raw_tensor
        # we can use log directly because the normal-space raw tensor came from neuralpp.exponentiation
        normalized_in_log_space_raw_tensor = normalized_in_normal_space_raw_tensor.log()
        return PyTorchLogTable(normalized_in_log_space_raw_tensor, self.batch)

    def randomize(self):
        self.raw_tensor = self.make_random_raw_tensor()

    def randomized_copy(self):
        leaf_log_tensor = self.make_random_raw_tensor()
        return PyTorchLogTable(leaf_log_tensor, self.batch)

    def make_random_raw_tensor(self):
        random_uniform_tensor = torch.rand(self.raw_tensor.shape)
        random_log_uniform_tensor = log_without_inf_non_differentiable(
            random_uniform_tensor
        )
        leaf_log_tensor = (
            random_log_uniform_tensor.clone().detach().requires_grad_(True)
        )
        return leaf_log_tensor

    def potentials_tensor(self):
        return self.raw_tensor.exp()



from itertools import chain

from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frames,
    to,
)
from neuralpp.inference.graphical_model.representation.model.model import (
    cross_entropy_for_datapoint,
)
from neuralpp.util.generic_sgd_learner import GenericSGDLearner, default_after_epoch


class GraphicalModelSGDLearner(GenericSGDLearner):
    """
    A GenericSGDLearner specialized to graphical models,
    making assumptions about datapoints being pairs of
    observation and query assignment batches,
    how to transfer them to a given device,
    and a loss function implementing cross entropy for them.
    """

    def __init__(
        self,
        model,
        data_loader,
        lr=1e-3,
        loss_decrease_tol=1e-3,
        after_epoch=default_after_epoch,
        debug=False,
        device=None,
        number_of_batches_between_updates=1,
        max_epochs_to_go_before_stopping_due_to_loss_decrease=1,
    ):

        super().__init__(
            model,
            data_loader,
            lr,
            loss_decrease_tol,
            device,
            number_of_batches_between_updates,
            debug,
            after_epoch,
            max_epochs_to_go_before_stopping_due_to_loss_decrease=max_epochs_to_go_before_stopping_due_to_loss_decrease,
        )

    def batch_to_device(self, batch, device):
        (observation_batch, query_assignment_batch) = batch
        observation_batch = to(observation_batch, device)
        query_assignment_batch = to(query_assignment_batch, device)
        return observation_batch, query_assignment_batch

    def get_number_of_datapoints_in_batch(self, batch):
        (observation_batch, query_assignment_batch) = batch
        return generalized_len_of_dict_frames(observation_batch, query_assignment_batch)

    def loss_function(self, batch):
        (observation_batch, query_assignment_batch) = batch
        return cross_entropy_for_datapoint(
            observation_batch, query_assignment_batch, self.model, self.debug
        )

    def get_parameters(self, model):
        return chain(*(f.pytorch_parameters() for f in model))

    def must_retain_graph_during_backpropagation(self):
        """
        Backpropagate with retain_graph because components may be included more
        than once due to factors sharing parameters.
        """
        return True



from neuralpp.inference.graphical_model.representation.frame.dict_data_loader import (
    DictDataLoader,
)
from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frame,
    is_frame,
)


class TupleDataLoader:
    def __init__(self, dataset_tuple, component_data_loader_maker):
        self.dataset_tuple = dataset_tuple
        self.data_loaders = [
            component_data_loader_maker(ds) for ds in self.dataset_tuple
        ]
        lengths = {len(data_loader) for data_loader in self.data_loaders}
        if len(lengths) != 1:
            raise DataLoadersMustHaveTheSameLength()
        (self.length,) = lengths

    def __iter__(self):
        return zip(*self.data_loaders)

    def __len__(self):
        return self.length


class DataLoadersMustHaveTheSameLength(BaseException):
    def __init__(self):
        super(DataLoadersMustHaveTheSameLength, self).__init__(
            "Data loaders must have the same length"
        )


def get_data_loader(observation_frame, query_assignment_frame, batch_size=100):
    try_breaking_data_into_batches = True
    dictionary = {**observation_frame, **query_assignment_frame}
    use_batches = try_breaking_data_into_batches and is_frame(dictionary)
    if use_batches:
        number_of_values = generalized_len_of_dict_frame(dictionary)
        batch_size = batch_size if number_of_values > batch_size else number_of_values
        frame_data_loader_maker = lambda frame: DictDataLoader(
            frame, batch_size=batch_size
        )
        data_loader = TupleDataLoader(
            (observation_frame, query_assignment_frame), frame_data_loader_maker
        )
    else:
        data_loader = [(observation_frame, query_assignment_frame)]
    return data_loader



import random

from neuralpp.inference.graphical_model.representation.random.random_model import (
    generate_model,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.graphical_model.variable_elimination import VariableElimination
from neuralpp.util.util import join, repeat


def generate_assignment_dict(variables):

    assert all(
        isinstance(v, IntegerVariable) for v in variables
    ), f"Random dataset generation is currently defined for {IntegerVariable.__name__} only"

    return {v: random.randrange(0, v.cardinality) for v in variables}


def condition(model, observation_dict):
    return [factor.condition(observation_dict) for factor in model]


def generate_dataset(
    model,
    number_of_sets_of_observed_and_query_variables,
    number_of_query_variables,
    number_of_observed_variables,
    number_of_observations_per_random_set_of_observed_and_query_variables,
    datapoints_per_observation,
):
    model_variables = {v for factor in model for v in factor.variables}
    dataset = []
    for i in range(number_of_sets_of_observed_and_query_variables):
        observation_dataset = (
            generate_dataset_given_number_of_observation_and_query_variables(
                model,
                model_variables,
                number_of_query_variables,
                number_of_observed_variables,
                number_of_observations_per_random_set_of_observed_and_query_variables,
                datapoints_per_observation,
            )
        )
        dataset = dataset + observation_dataset
    return dataset


def generate_dataset_given_number_of_observation_and_query_variables(
    model,
    model_variables,
    number_of_query_variables,
    number_of_observed_variables,
    number_of_observations_per_random_set_of_observed_and_query_variables,
    datapoints_per_observation,
):

    observed_variables = set(
        random.sample(model_variables, number_of_observed_variables)
    )
    query_variables = random.sample(
        list(model_variables - observed_variables), number_of_query_variables
    )

    return generate_dataset_given_observation_and_query_variables(
        model,
        observed_variables,
        query_variables,
        number_of_observations_per_random_set_of_observed_and_query_variables,
        datapoints_per_observation=datapoints_per_observation,
    )


def generate_dataset_given_observation_and_query_variables(
    model,
    observed_variables,
    query_variables,
    number_of_observations,
    datapoints_per_observation,
):
    dataset = []
    for i in range(number_of_observations):
        observation_dict = generate_assignment_dict(observed_variables)
        dataset_for_observation_dict = (
            generate_dataset_given_observation_dict_and_query_variables(
                datapoints_per_observation, model, observation_dict, query_variables
            )
        )
        dataset.extend(dataset_for_observation_dict)
    return dataset


def generate_dataset_given_observation_dict_and_query_variables(
    datapoints_per_observation, model, observation_dict, query_variables
):
    conditioned_model = condition(model, observation_dict)
    query_distribution = (
        VariableElimination().run(query_variables, conditioned_model).atomic_factor()
    )
    observation_dataset = repeat(
        datapoints_per_observation,
        lambda: (observation_dict, query_distribution.single_sample_assignment_dict()),
    )
    # print(f"observation_dict: {observation_dict}")
    # print(f"conditioned_model: {conditioned_model}")
    # print(f"query_variables: {query_variables}")
    # print(f"query_distribution: {query_distribution}")
    # observation_dataset_string = join(observation_dataset, '\n')
    # print(f"observation_dataset:\n{observation_dataset_string}")
    return observation_dataset


if __name__ == "__main__":
    model = generate_model(number_of_factors=6, number_of_variables=4, cardinality=3)
    print("Model:")
    print(join(model, "\n"))
    print()

    ds = generate_dataset(
        model=model,
        number_of_sets_of_observed_and_query_variables=5,
        number_of_query_variables=random.randint(0, 2),
        number_of_observed_variables=2,
        number_of_observations_per_random_set_of_observed_and_query_variables=1,
        datapoints_per_observation=10,
    )
    print()
    print("Generated dataset:")
    print(join(ds, "\n"))



import torch
from neuralpp.util.util import has_len, cartesian_prod_2d, repeat_first_dimension_with_expand, \
    repeat_interleave_first_dimension


def is_frame(dictionary):
    return all(has_len(v) for v in dictionary.values())


def generalized_len_of_dict_frames(*dict_frames):
    if len(dict_frames) == 0:
        raise ThereShouldBeAtLeastOneDictFrame()
    set_of_lengths = {generalized_len_of_dict_frame(frame) for frame in dict_frames}
    if len(set_of_lengths) != 1:
        raise DictFramesShouldAllHaveTheSameLength()
    (length,) = set_of_lengths
    return length


def generalized_len_of_dict_frame(dict_frame):
    set_of_lengths = compute_set_of_lengths(dict_frame)
    if len(set_of_lengths) != 1:
        raise DictionaryValuesShouldAllHaveTheSameLength()
    (length,) = set_of_lengths
    return length


# def len_of_expanded_dict_frame(dict_frame):
#     """
#     Same as generalized_len_of_dict_frame, but considering
#     that univalues will be expanded to the length of multivalues.
#     """
#     set_of_lengths = compute_set_of_lengths(dict_frame)
#     if len(set_of_lengths) == 1:
#         length = next(iter(set_of_lengths))
#     elif len(set_of_lengths) == 2 and 1 in set_of_lengths:
#         length = next(l for l in set_of_lengths if l != 1)
#     else:
#         raise DictionaryValuesShouldAllHaveTheSameLength()
#     return length


def expand_univalues_in_dict_frame(tensor_dict_frame):
    """
    Assumes tensor_dict_frame has tensor values which are either univalues, or multivalues of the same length.
    Returns a new dict frame with the same multivalues and univalues expanded to same length as multivalues.
    """
    multivalue_lengths = compute_set_of_multivalue_lengths(tensor_dict_frame)
    if len(multivalue_lengths) == 0:
        return tensor_dict_frame
    elif len(multivalue_lengths) > 1:
        raise DictFramesShouldAllHaveTheSameLength()
    else:
        multivalue_length = next(iter(multivalue_lengths))
        return \
            {
                variable: (
                    value.expand(multivalue_length, *((-1,)*value.dim()))
                    if not variable.is_multivalue(value)
                    else value
                )
                for variable, value in tensor_dict_frame.items()
            }


def compute_set_of_lengths(dict_frame):
    """
    Returns the set of value lengths of values in dict_frame.
    """
    if len(dict_frame) == 0:
        raise DictionaryShouldHaveAtLeastOneItem()
    set_of_lengths = {variable.value_len(value) for variable, value in dict_frame.items()}
    return set_of_lengths


def compute_set_of_multivalue_lengths(dict_frame):
    """
    Returns the set of lengths of multivalues in dict_frame.
    """
    return {variable.value_len(value)
            for variable, value in dict_frame.items()
            if variable.is_multivalue(value)}


def number_of_equal_values_in_dict_frames(dict_frame1, dict_frame2):
    if dict_frame1.keys() != dict_frame2.keys():
        raise DictionariesShouldHaveTheSameKeys()
    assert_values_are_tensors(dict_frame1)
    assert_values_are_tensors(dict_frame2)
    column_comparisons = [dict_frame1[k].eq(dict_frame2[k]) for k in dict_frame1]
    comparisons_matrix = torch.stack(column_comparisons, dim=1).bool()
    number_of_equal_rows = comparisons_matrix.all(dim=1).sum().item()
    return number_of_equal_rows


def assert_values_are_tensors(dict_frame1):
    assert all(
        isinstance(v, torch.Tensor) for v in dict_frame1.values()
    ), "number_of_equal_values_in_dict_frames currently implemented for Tensor values only"


def to(dict_frame, device):
    if device is not None:
        return {v: to_if_tensor(data, device) for v, data in dict_frame.items()}
    else:
        return dict_frame


def to_if_tensor(obj, device):
    if isinstance(obj, torch.Tensor):
        return obj.to(device)
    else:
        return obj


def featurize_dict_frame(dict_frame):
    return {variable: variable.featurize(value) for variable, value in dict_frame.items()}


def make_cartesian_features_dict_frame(variables):
    if len(variables) > 0:
        free_cardinalities = [torch.arange(fv.cardinality) for fv in variables]
        free_assignments = cartesian_prod_2d(free_cardinalities)
        cartesian_free_features_dict_frame = {variable: free_assignments[:, i]
                                              for i, variable in enumerate(variables)}
    else:
        cartesian_free_features_dict_frame = {}
    return cartesian_free_features_dict_frame


def concatenate_into_single_tensor(dict_frame):
    if len(dict_frame) > 0:
        tensor_2d = concatenate_non_empty_dict_frame_into_single_2d_tensor(dict_frame)
    else:
        tensor_2d = torch.ones(1, 0)
    batch = any(variable.is_multivalue(value) for variable, value in dict_frame.items())
    if batch:
        tensor = tensor_2d
    else:
        tensor = tensor_2d[0]
    return tensor


def concatenate_non_empty_dict_frame_into_single_2d_tensor(expanded_tensor_dict_frame):
    """
    Given an ordered non-empty dictionary frame with all multivalue values of same length
    (but some possible univalues),
    returns a 2D tensor where element (i,j) is
    the i-th value of the j-th variable.
    """
    expanded_dict_frame_with_2d_tensor_values = convert_tensor_values_to_at_least_two_dimensions(
        expanded_tensor_dict_frame
    )
    conditioning_tensor = torch.cat(tuple(expanded_dict_frame_with_2d_tensor_values.values()), dim=1)
    return conditioning_tensor


def convert_tensor_values_to_at_least_two_dimensions(dict_frame):
    return {
        k: unsqueeze_if_needed_for_at_least_two_dimensions(v)
        for k, v in dict_frame.items()
    }


def unsqueeze_if_needed_for_at_least_two_dimensions(tensor):
    assert isinstance(tensor, torch.Tensor), "Must be tensor argument."
    while tensor.dim() < 2:
        tensor = tensor.unsqueeze(1)
    return tensor


def cartesian_product_of_tensor_dict_frames(dict_frame1, dict_frame2):
    n = generalized_len_of_dict_frame(dict_frame1) \
        if len(dict_frame1) > 0 \
        else 1
    m = generalized_len_of_dict_frame(dict_frame2) \
        if len(dict_frame2) > 0 \
        else 1
    big_endian_dict_frame1 = repeat_interleave_dict_frame(dict_frame1, m)
    little_endian_dict_frame2 = repeat_dict_frame(dict_frame2, n)
    cartesian_product_dict_frame = {**big_endian_dict_frame1, **little_endian_dict_frame2}
    return cartesian_product_dict_frame


def repeat_dict_frame(dict_frame, n):
    """
    Returns new dict frame with values equal to result of expanding the first dimension of every tensor value in
    the original (using util.repeat_first_dimension_with_expand for each value).
    """
    return {variable: repeat_first_dimension_with_expand(value, n) for variable, value in dict_frame.items()}


def repeat_interleave_dict_frame(dict_frame, n):
    """
    Returns new dict frame with values equal to repeat-interleaving the elements of the first dimenson
    (using util.repeat_interleave_first_dimension for each value).
    """
    return {variable: repeat_interleave_first_dimension(value, n) for variable, value in dict_frame.items()}


class DictionaryShouldHaveAtLeastOneItem(BaseException):
    def __init__(self):
        super(DictionaryShouldHaveAtLeastOneItem, self).__init__(
            "Dictionary should have at least one item"
        )


class DictionaryValuesShouldAllHaveTheSameLength(BaseException):
    def __init__(self):
        super(DictionaryValuesShouldAllHaveTheSameLength, self).__init__(
            "Dictionary values should all have the same length"
        )


class DictionaryValuesShouldAllEitherHaveLengthOneOrSomeOtherSharedLength(BaseException):
    def __init__(self):
        super(DictionaryValuesShouldAllEitherHaveLengthOneOrSomeOtherSharedLength, self).__init__(
            "Dictionary values should all have either length 1 or some other shared length"
        )


class ThereShouldBeAtLeastOneDictFrame(BaseException):
    def __init__(self):
        super(ThereShouldBeAtLeastOneDictFrame, self).__init__(
            "There should be at least one dict frame"
        )


class DictFramesShouldAllHaveTheSameLength(BaseException):
    def __init__(self):
        super(DictFramesShouldAllHaveTheSameLength, self).__init__(
            "Dict frames should all have the same length"
        )


class DictionariesShouldHaveTheSameKeys(BaseException):
    def __init__(self):
        super(DictionariesShouldHaveTheSameKeys, self).__init__(
            "Dictionaries should have the same keys"
        )



import functools
from math import prod

import torch
from neuralpp.inference.graphical_model.representation.factor.atomic_factor import (
    AtomicFactor,
)
from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frame, featurize_dict_frame, expand_univalues_in_dict_frame,
    concatenate_non_empty_dict_frame_into_single_2d_tensor, repeat_dict_frame, repeat_interleave_dict_frame,
    cartesian_product_of_tensor_dict_frames, concatenate_into_single_tensor, make_cartesian_features_dict_frame,
    compute_set_of_multivalue_lengths,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)
from neuralpp.util import util
from neuralpp.util.util import find, join, is_iterable, expand_into_batch, cartesian_prod_2d, \
    cartesian_product_of_two_tensors, dict_slice


class NeuralFactor(AtomicFactor):
    """
    Factor based on a neural net, which must conform to the following interface:
    probability_tensor_indexed_by_output_variable_values = neural_net(features),
        for features the concatenation of feature vectors of all variables given full assignment
        and construction-time conditioning
    neural_net.randomize(): randomizes weights in-place
    neural_net.randomized_copy(): provides a copy of neural_net with randomized weights
    neural_net(concatenation of feature vectors of all variables, input ones and output one, given full assignment)

    For convenience, the last layer of the neural network may also just return
    the probabilities of the last n - 1 values of output_variable,
    in which case the 0-th probability will be computed as 1 - sum of the others.
    """

    def __init__(
            self, neural_net, input_variables, output_variable, conditioning_dict={}
    ):

        super().__init__(
            self.non_conditioned_variables(
                input_variables, output_variable, conditioning_dict.keys()
            )
        )

        assert isinstance(
            output_variable, DiscreteVariable
        ), f"{NeuralFactor.__name__} requires output variable {output_variable} to be {DiscreteVariable.__name__}"
        assert (
                output_variable.cardinality is not None
        ), f"{NeuralFactor.__name__} requires output variable {output_variable} to have a well-defined cardinality, " \
           f"but {output_variable} has none"

        self.input_variables = input_variables
        self.free_input_variables = [
            v for v in input_variables if v not in conditioning_dict
        ]
        self.output_variable = output_variable
        self.conditioning_dict = conditioning_dict
        self.neural_net = neural_net

    @staticmethod
    def non_conditioned_variables(
            input_variables, output_variable, conditioning_variables
    ):
        return [
            v
            for v in input_variables + [output_variable]
            if v not in conditioning_variables
        ]

    def pytorch_parameters(self):
        return self.neural_net.parameters()

    def call_after_validation_MODEL(self, assignment_dict):
        return self.to_table_factor_if_output_variable_is_conditioned_MODEL(assignment_dict).table.tensor
        # TODO: change underlying table from log to normal space

    def call_after_validation(self, assignment_dict, assignment_values):
        probabilities = self.probabilities_from_assignment_dict(assignment_dict)
        output_value = assignment_dict[self.output_variable]

        # the following code deals with the possibility that probabilities and output_value
        # differ in the length (number of values / batchness).
        # We could have normalized the assignment_dict to have all values of the same length,
        # but that might make the neural network run many times over the same input
        # (for example, if the input was univalue and the output multivalue,
        # then we would torch.expand the inputs and the neural network would apply
        # on many identical multiple roles).
        # So what we do instead is not to normalize the assignment, let probabilities
        # be computed in the most economical form given inputs,
        # and then deal with the possibility that probabilities and output_value
        # need to be adjusted to each other at this level.

        assert probabilities.dim() in {1, 2}
        assert isinstance(output_value, int) or is_iterable(output_value)

        if is_iterable(output_value):
            if probabilities.dim() == 1:
                probabilities = probabilities.expand(len(output_value), -1)
            if not isinstance(output_value, torch.Tensor):
                output_value = torch.tensor(output_value)
            probability = probabilities.gather(dim=1, index=output_value.unsqueeze(1)).squeeze()
        else:
            if probabilities.dim() == 1:
                probability = probabilities[output_value]
            else:
                probability = probabilities[:, output_value]

        return probability

    def probabilities_from_assignment_dict(self, assignment_dict):
        neural_net_input = self.neural_net_input_from_assignment_dict(assignment_dict)
        probabilities = self.output_probabilities(neural_net_input)
        return probabilities

    def neural_net_input_from_assignment_dict(self, assignment_dict):
        self.check_assignment_dict_is_complete(assignment_dict)
        self.check_assignment_dict_does_not_contradict_conditioning_dict(assignment_dict)

        assignment_and_conditioning_dict = util.union_of_dicts(
            assignment_dict, self.conditioning_dict
        )

        # featurize
        # check lengths
        # obtain multivalues length
        # if there are multivalues
        #     expand univalues into 2D tensors
        #     concatenate along dimension 1
        # else
        #     concatenate along dimension 0

        multivalue_lengths = compute_set_of_multivalue_lengths(assignment_and_conditioning_dict)

        if len(multivalue_lengths) > 1:
            raise Exception(f"neural factor received multivalue inputs of different lengths: {multivalue_lengths}")

        tuple_of_featurized_value_tensors = tuple(
            v.featurize(assignment_and_conditioning_dict[v])
            for v in self.input_variables
        )

        def variable_value_featurized():
            return (
                (v, assignment_and_conditioning_dict[v], fvt)
                for v, fvt in zip(self.input_variables, tuple_of_featurized_value_tensors)
            )

        there_are_multivalues = len(multivalue_lengths) == 1

        if there_are_multivalues:
            batch_size = next(iter(multivalue_lengths))
            tuple_of_featurized_value_tensors = tuple(
                expand_into_batch(fvt, batch_size) if not v.is_multivalue(value) else fvt
                for v, value, fvt in variable_value_featurized()
            )

        dim_to_concatenate = 1 if there_are_multivalues else 0
        neural_net_input = torch.cat(tuple_of_featurized_value_tensors, dim=dim_to_concatenate)

        return neural_net_input

    def output_probabilities(self, neural_net_input):
        try:
            probabilities = self.neural_net(neural_net_input)
        except RuntimeError as e:
            if "size mismatch" in str(e):
                raise Exception(
                    f"Size mismatch exception when applying neural network {self.neural_net}, possibly due to mismatching feature vector (length {len(neural_net_input)}) and neural net input layer size. Original exception was: {e}"
                )
            else:
                raise e
        # using shape[-1] takes care of both cases of probabilities being a batch or not.
        assert (
                probabilities.shape[-1] == self.output_variable.cardinality
        ), f"Neural net {self.neural_net} output must have the same size as output variable '{self.output_variable}' cardinality {self.output_variable.cardinality}, but has size {len(probabilities)} instead"
        return probabilities

    def condition_on_non_empty_dict(self, assignment_dict):
        # self.check_conditioning_is_on_factors_variables_only(assignment_dict)
        new_conditioning_dict = util.union_of_dicts(
            assignment_dict, self.conditioning_dict
        )
        return NeuralFactor(
            self.neural_net,
            self.input_variables,
            self.output_variable,
            new_conditioning_dict,
        )

    def check_conditioning_is_on_factors_variables_only(self, assignment_dict):
        extra_variable = util.find(
            assignment_dict.keys(), lambda v: v not in self.variables
        )
        if extra_variable:
            raise Exception(
                f"Factor conditioned on {extra_variable} but that is not one of its variables {self.variables}. Factor is {self}"
            )

    def randomize(self):
        self.neural_net.randomize()

    def randomized_copy(self):
        return NeuralFactor(
            self.neural_net.randomized_copy(),
            self.input_variables,
            self.output_variable,
            self.conditioning_dict,
        )

    def mul_by_non_identity(self, other):
        return other * self.table_factor
        # order matters if other is also neural, because tables do not know how to multiply by neural factors.
        # This way, if other is also neural, it will itself be converted to a table.

    def sum_out_variable(self, variable):
        # TODO: This can be optimized for the case in which variable is the output variable (result is uniform factor
        # on free variables).
        return self.table_factor.sum_out_variable(variable)

    def argmax(self):
        return self.table_factor.argmax()

    def normalize(self):
        if self.variables == [self.output_variable]:
            return self
        else:
            return self.table_factor.normalize()

    def sample(self):
        return self.table_factor.sample()

    def single_sample(self):
        return self.table_factor.single_sample()

    @property
    @functools.lru_cache(1)
    def table_factor(self):
        if self.output_variable in self.conditioning_dict:
            return self.to_table_factor_if_output_variable_is_conditioned()
        else:
            return self.to_table_factor_if_output_variable_is_not_conditioned()

    def to_table_factor_if_output_variable_is_conditioned(self):
        table_factor = self.to_table_factor_if_output_variable_is_not_conditioned()
        return table_factor.condition({self.output_variable: self.conditioning_dict[self.output_variable]})

    def to_table_factor_if_output_variable_is_not_conditioned(self):
        probabilities_tensor = self.compute_probability_tensor()
        resulting_factor = self.make_table_factor_for_free_and_output_variables(probabilities_tensor)
        return resulting_factor

    def compute_probability_tensor(self):
        return self.compute_probabilities_tensor_for(self.conditioning_dict)

    def compute_probabilities_tensor_for(self, assignment_dict, no_free_variables = False):
        """
        Returns the probabilities tensor produced by the neural net for a given assignment_dict.
        The probabilities tensor will have an optional batch dimension (determined by whether any
        values in the conditioning dict is multivalue OR if there are free input variables that will be automatically
        filled in with all assignments), followed by other dimensions in the order of self.input_variables.

        If the caller knows there are no free variables, they can pass no_free_variables = True for optimization.
        In this case, the assignment_dict keys are required to be in the same order as self.input_variables.

        If there are free variables, they are assumed to be IntegerVariables.
        """
        relevant_conditioning_dict_frame = dict_slice(assignment_dict, self.input_variables)
        featurized_conditioning_dict_frame = featurize_dict_frame(relevant_conditioning_dict_frame)
        expanded_featurized_conditioning_dict_frame = expand_univalues_in_dict_frame(featurized_conditioning_dict_frame)

        if no_free_variables:
            all_inputs_dict_frame = expanded_featurized_conditioning_dict_frame
        else:
            all_inputs_dict_frame = self.complete_with_free_variables(expanded_featurized_conditioning_dict_frame)

        all_inputs_tensor = concatenate_into_single_tensor(all_inputs_dict_frame)
        probabilities_tensor = self.neural_net(all_inputs_tensor)
        return probabilities_tensor

    def complete_with_free_variables(self, expanded_featurized_conditioning_dict_frame):
        free_input_variables = self.input_variables - expanded_featurized_conditioning_dict_frame.keys()
        cartesian_free_features_dict_frame = make_cartesian_features_dict_frame(free_input_variables)
        all_inputs_dict_frame = cartesian_product_of_tensor_dict_frames(expanded_featurized_conditioning_dict_frame,
                                                                        cartesian_free_features_dict_frame)
        return all_inputs_dict_frame

    def make_cartesian_features_dict_frame(variables):
        if len(variables) > 0:
            free_cardinalities = [torch.arange(fv.cardinality) for fv in variables]
            free_assignments = cartesian_prod_2d(free_cardinalities)
            cartesian_free_features_dict_frame = {variable: free_assignments[:, i]
                                                  for i, variable in enumerate(variables)}
        else:
            cartesian_free_features_dict_frame = {}
        return cartesian_free_features_dict_frame

    def make_table_factor_for_free_and_output_variables(self, probabilities_tensor):
        number_of_batch_rows = probabilities_tensor.numel() // self.number_of_probabilities_per_batch_row
        batch = number_of_batch_rows != 1
        if batch:
            probabilities_shape = (
                number_of_batch_rows,
                *self.non_batch_shape_including_output_variable,
            )
        else:
            probabilities_shape = self.non_batch_shape_including_output_variable
        probabilities_tensor_in_right_shape = probabilities_tensor.reshape(probabilities_shape)
        resulting_factor = PyTorchTableFactor(
            self.free_input_variables + [self.output_variable],
            probabilities_tensor_in_right_shape,
            log_space=True,
            batch=batch,
        )
        return resulting_factor

    @property
    @functools.lru_cache(1)
    def non_batch_shape_including_output_variable(self):
        return tuple(v.cardinality for v in self.free_input_variables) + (self.output_variable.cardinality,)

    @property
    @functools.lru_cache(1)
    def number_of_probabilities_per_batch_row(self):
        return prod(self.non_batch_shape_including_output_variable)

    def _check_whether_free_input_variables_can_be_enumerated(self):
        if any(v.assignments is None for v in self.free_input_variables):
            raise Exception(
                f"{type(self)}: input variables that are free at inference time "
                f"must have method 'assignments'"
            )

    def __repr__(self):
        result = (
                repr(self.neural_net)
                + f" on {join(self.input_variables)} -> {self.output_variable}"
        )
        if self.conditioning_dict:
            result += f" conditioned on {self.conditioning_dict}"
        return result

    def assignment_dict_does_not_contradict_conditioning(self, assignment_dict):
        for var, val in assignment_dict.items():
            if var in self.conditioning_dict and val != self.conditioning_dict[var]:
                return False
        return True

    def check_assignment_dict_is_complete(self, assignment_dict):
        assert all(
            v in assignment_dict for v in self.variables
        ), f"{NeuralFactor.__name__} received a non-complete assignment. It is missing the neural factor's variable '{find(self.variables, lambda v: v not in assignment_dict)}'. The assignment is {assignment_dict}"

    def check_assignment_dict_does_not_contradict_conditioning_dict(
            self, assignment_dict
    ):
        assert self.assignment_dict_does_not_contradict_conditioning(
            assignment_dict
        ), f"Assignment dict {assignment_dict} contradicts neural factor conditioning {self.conditioning_dict}"



from itertools import islice

import torch
from neuralpp.util.pickle_cache import pickle_cache
from neuralpp.util.util import get_or_put, go_up_until_we_have_subdirectory
from torchvision import datasets, transforms
from tqdm import tqdm


read_mnist_from_original_files = False


def read_mnist(max_datapoints=None):
    go_up_until_we_have_subdirectory("data")
    return pickle_cache(
        lambda: read_mnist_no_cache(max_datapoints),
        f"data/cache/indexed_mnist_max_points={max_datapoints}.pkl",
        read_mnist_from_original_files,
    )


def read_mnist_no_cache(max_datapoints=None):
    go_up_until_we_have_subdirectory("data")
    """Return a map from phase to a map from digit to a list of its images: images_for_digit_by_phase[phase][digit]"""
    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,)),
        ]
    )
    phases = {"train", "test"}
    loader = {}
    images_by_digits = {}
    for p in phases:
        mnist_dataset = datasets.MNIST(
            "data", train=(p == "train"), download=True, transform=transform
        )
        loader[p] = torch.utils.data.DataLoader(mnist_dataset, batch_size=1)
        images_by_digits[p] = {}
        print(f"Indexing {p}ing MNIST dataset...")
        i = 0
        for image_singleton_batch, digit in tqdm(loader[p]):
            image = image_singleton_batch[0][
                0
            ]  # get single row of batch and single color
            images_of_digit = get_or_put(images_by_digits[p], digit.item(), [])
            images_of_digit.append(image)
            i += 1
            if max_datapoints is not None and i >= max_datapoints:
                break
    return images_by_digits


def evaluate_digits(net, get_images_of_digit, max_images_per_digit=100):
    print()
    for digit in range(10):
        number_of_correct_answers = 0
        images = get_images_of_digit(digit)[:max_images_per_digit]
        for image in images:
            prediction = net(image).argmax()
            if prediction == digit:
                number_of_correct_answers += 1
        print(f"Accuracy for {digit}: {number_of_correct_answers/len(images):.2f}")
    print()
    input("Press Enter for continuing")


def show_images_and_labels(n_rows, n_cols, images, labels):
    from matplotlib import pyplot as plt

    plt.figure()
    n_positions = n_rows * n_cols
    for i, (image, label) in enumerate(
        islice(zip(images, labels), n_positions), start=1
    ):
        plt.subplot(n_rows, n_cols, i)
        plt.tight_layout()
        plt.imshow(image, cmap="gray", interpolation="none")
        plt.title(label)
        plt.xticks([])
        plt.yticks([])
        i += 1
    plt.show()



import torch

from neuralpp.inference.graphical_model.representation.factor.continuous.continuous_factor import ContinuousFactor
from neuralpp.util import util
from neuralpp.util.util import join


class ContinuousInternalParameterlessFactor(ContinuousFactor):
    """
    A specialization of ContinuousFactor that does not use any internal parameters.
    """

    def __init__(self, variables, conditioning_dict=None):
        super().__init__(variables, conditioning_dict)

        # randomization does not apply since there are no internal parameters.
        self.randomize = None
        self.randomized_copy = None

    def __eq__(self, other):
        """
        Compares factors by checking if they are both NormalFactors and have the same variables and conditioning
        """
        return isinstance(other, type(self)) \
               and self.variables == other.variables \
               and self.conditioning_dict == other.conditioning_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{type(self)}(" + join(self.variables) + ")" + \
               (" given " + self.conditioning_dict if self.conditioning_dict else "")



import math

import torch
from neuralpp.util.tensor_mixed_radix import MaxValueException, TensorMixedRadix


def test_tensor_mixed_radix_representation():
    values = [0, 1, 2, 3]
    radices = (2, 2)
    expected = [(0, 0), (0, 1), (1, 0), (1, 1)]
    run_tensor_mixed_radix_representation(values, radices, expected)

    values = [0, 1, 2, 3]
    radices = (4, 4)
    expected = [(0, 0), (0, 1), (0, 2), (0, 3)]
    run_tensor_mixed_radix_representation(values, radices, expected)

    values = [0, 1, 2, 3]
    radices = (4, 2)
    expected = [(0, 0), (0, 1), (1, 0), (1, 1)]
    run_tensor_mixed_radix_representation(values, radices, expected)

    values = [6, 7]
    radices = (4, 2)
    expected = [(3, 0), (3, 1)]
    run_tensor_mixed_radix_representation(values, radices, expected)

    values = [7, 8, 9, 10]
    radices = (2, 4, 2)
    expected = [(0, 3, 1), (1, 0, 0), (1, 0, 1), (1, 1, 0)]
    run_tensor_mixed_radix_representation(values, radices, expected)

    values = [7, 8, 16, 10]  # 16 is greater than max value 15
    radices = (2, 4, 2)
    violating_value = 16
    check_expected_exception(
        values, radices, max_value_exception(violating_value, radices)
    )

    values = [1]  # 1 is greater than max value 0
    radices = (1,)
    violating_value = 1
    check_expected_exception(
        values, radices, max_value_exception(violating_value, radices)
    )

    values = [1]  # 1 is greater than max value 0
    radices = (1, 1, 1)
    violating_value = 1
    check_expected_exception(
        values, radices, max_value_exception(violating_value, radices)
    )

    values = [1]  # 1 is greater than max value 0
    radices = ()
    violating_value = 1
    check_expected_exception(
        values, radices, max_value_exception(violating_value, radices)
    )


def max_value_exception(violating_value, radices):
    return MaxValueException(violating_value, math.prod(radices) - 1)


def check_expected_exception(values, radices, expected_exception):
    try:
        mixed_radix = TensorMixedRadix(radices)
        mixed_radix.representation(torch.tensor(values))
        raise AssertionError(f"Should have thrown {expected_exception}")
    except Exception as e:
        assert e == expected_exception


def run_tensor_mixed_radix_representation(values, radices, expected):
    mixed_radix = TensorMixedRadix(radices)
    actual = mixed_radix.representation(torch.tensor(values))
    print(f"Indices of {values} with radices {radices}: {actual}")
    assert torch.allclose(actual, torch.tensor(expected, dtype=torch.int))



import torch
from neuralpp.util import util
from neuralpp.util.util import join, pairwise, vararg_or_array
from torch import nn
from torch.nn import Module


class MLP(Module):
    """
    A multi-layer perceptron composed of a configurable number of linear layers followed by sigmoid transformations.
    The last layer output elements will numbers between 0 and 1.
    They do not necessarily sum up to 1; each must be considered a separate probability.
    """

    def __init__(self, *all_layer_sizes):
        super(MLP, self).__init__()
        self.layer_sizes = vararg_or_array(all_layer_sizes)
        assert len(self.layer_sizes) > 0
        assert all(
            isinstance(layer_size, int) for layer_size in self.layer_sizes
        ), f"Layer sizes must be ints but got a layer size value equal to {util.find(self.layer_sizes, lambda e: not isinstance(e, int))}"
        assert all(
            layer_size > 0 for layer_size in self.layer_sizes
        ), f"MLP layers must have sizes greater than zero, but were {all_layer_sizes}"
        self.linear_transformations = nn.ModuleList(
            [torch.nn.Linear(i, o) for i, o in pairwise(self.layer_sizes)]
        )
        self.last_layer_index = len(self.linear_transformations) - 1

    def forward(self, x):
        assert isinstance(x, torch.Tensor)
        if not torch.is_floating_point(x):
            x = x.float()
        if x.shape[-1] != self.layer_sizes[0]:
            x = x.reshape(-1, self.layer_sizes[0])
        for i, linear_transformation in enumerate(self.linear_transformations):
            transformation = linear_transformation(x)

            # print("----------------------")
            # print(rounded_list(x))
            # print("*")
            # print(rounded_list(linear_transformation.weight))
            # print("+")
            # print(rounded_list(linear_transformation.bias))
            # print("=")
            # print(rounded_list(transformation))

            if i == self.last_layer_index:
                x = torch.nn.functional.softmax(transformation, dim=-1)
                # print("given to softmax ->")
                # print(rounded_list(x))
            else:
                x = torch.sigmoid(transformation)
                # print("given to sigmoid ->")
                # print(rounded_list(x))
        return x

    def randomize(self):
        for layer in self.linear_transformations:
            # we pick a random number in [-5.0, 5.0] because that is the range needed
            # to ensure sigmoid produces values from near 0 to near 1,
            # divided by number of inputs so it does not saturate when there are many inputs
            n_inputs = layer.weight.shape[1] + 1
            uniform_range = 5.0 / n_inputs
            torch.nn.init.uniform_(layer.weight, -uniform_range, uniform_range)
            torch.nn.init.uniform_(layer.bias, -uniform_range, uniform_range)

            # print("n_inputs:", n_inputs)
            # print("Randomized layer.weight:", layer.weight)
            # print("Randomized layer.bias:", layer.bias)

    def randomized_copy(self):
        result = MLP(self.layer_sizes)
        result.randomize()
        return result

    def __repr__(self):
        return f"MLP({join(self.layer_sizes)})"



from typing import Any

import torch


class Variable:
    def __eq__(self, other) -> bool:
        self._not_implemented("__eq__")

    def __hash__(self):
        self._not_implemented("__hash__")

    def __repr__(self) -> str:
        self._not_implemented("__repr__")

    def featurize(self, value) -> torch.Tensor:
        self._not_implemented("featurize")

    def is_multivalue(self, value: Any) -> bool:
        self._not_implemented("is_multivalue")

    def value_len(self, value: Any) -> int:
        """
        Returns the length of the value if it is a multivalue, or 1 otherwise.
        """
        return self.multivalue_len(value) if self.is_multivalue(value) else 1

    def multivalue_len(self, value: Any) -> int:
        self._not_implemented("multivalue_len")

    def _not_implemented(self, name):
        error = NotImplementedError(f"{name} not implemented for {type(self)}")
        raise error



from neuralpp.inference.graphical_model.representation.factor.atomic_factor import (
    AtomicFactor,
)
from neuralpp.inference.graphical_model.representation.table.table_util import (
    index_of,
    permutation_from_to,
    shape,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)
from neuralpp.util.discrete_sampling import discrete_sample
from neuralpp.util.util import join


class TableFactor(AtomicFactor):
    """
    TableFactor is a Factor based on a Table.
    Like Table, a TableFactor may be a batch or not.
    """

    def __init__(self, variables, table):
        super().__init__(variables)
        self.table = table

    def new_instance(self, variables, table):
        return type(self)(variables, table)

    @property
    def batch(self):
        return self.table.batch

    def assignments(self):
        return DiscreteVariable.assignments_product(self.variables)

    def condition_on_non_empty_dict(self, assignment_dict):
        non_batch_slice_coordinates = tuple(
            assignment_dict[v] if v in assignment_dict else slice(None)
            for v in self.variables
        )
        non_conditioned_variables = [
            v
            for v in self.variables
            if v not in assignment_dict or isinstance(assignment_dict[v], slice)
        ]
        conditioned_table = self.table.slice(non_batch_slice_coordinates)
        return self.new_instance(non_conditioned_variables, conditioned_table)

    def call_after_validation(self, assignment_dict, assignment_values):
        return self.table[assignment_values]

    def mul_by_non_identity(self, other):
        """Multiplies factors so that (f1 * f2)(assignment) = f1(assignment)*f2(assignment)"""

        # TableFactor only knows how to multiply by another TableFactor.
        # If other is not TableFactor, we ask that it handle the multiplication.
        if not isinstance(other, TableFactor):
            return other * self

        """
        Implementation rationale:
        Let 'common' be the shared variables between the two factors, and e1 and e2
        the variables exclusive to each representation respectively.

        We want to compute f(e1, e2, common) = f1(e1, common) * f2(e2, common)
        Such multiplication operation is not readily available, however.
        Instead, we use component-wise multiplication by computing:
        for all e1, e2, common:
        f1'(e1, e2, common) = f1(e1, common)
        f2'(e1, e2, common) = f2(e2, common)
        f(e1, e2, common) = f1'(e1, e2, common) * f2'(e1, e2, common)

        We compute f1' and f2' by using Table.expand

        Note that things work transparently if tables are batches.
        """

        common = list(set(self.variables) & set(other.variables))

        (
            f1_e1_common_table,
            e1,
        ) = self.get_permuted_table_with_selected_variables_moved_to_the_end(
            self, common
        )
        (
            f2_e2_common_table,
            e2,
        ) = self.get_permuted_table_with_selected_variables_moved_to_the_end(
            other, common
        )

        result_variables = e1 + e2 + common

        f1p_table = f1_e1_common_table.expand(
            shape_to_be_inserted=shape(e2), dim=len(e1)
        )
        f2p_table = f2_e2_common_table.expand(shape_to_be_inserted=shape(e1), dim=0)
        result_table = f1p_table * f2p_table

        result = self.new_instance(result_variables, result_table)

        return result

    @staticmethod
    def get_permuted_table_with_selected_variables_moved_to_the_end(
        factor, selected_variables
    ):
        other_variables = [v for v in factor.variables if v not in selected_variables]
        variables_in_desired_order = other_variables + selected_variables
        permuted_table = factor.table.permute(
            index_of(variables_in_desired_order, factor.variables)
        )
        return permuted_table, other_variables

    def sum_out_variable(self, variable):
        result_variables = [v for v in self.variables if v != variable]
        index_of_variable = self.variables.index(variable)
        result_table = self.table.sum_out(index_of_variable)
        result = self.new_instance(result_variables, result_table)
        return result

    def argmax(self):
        indices = self.table.argmax()
        if len(self.variables) == 1:  # if there is a single variable, indices is 1D
            assignment_getter = (
                lambda var_index: indices
            )  # regardless of being batch or not
        else:
            assignment_getter = (
                lambda var_index: indices[:, var_index]
                if self.batch
                else indices[var_index]
            )
        assignment_dict = {
            v: assignment_getter(var_index)
            for var_index, v in enumerate(self.variables)
        }
        return assignment_dict

    def normalize(self):
        return self.new_instance(self.variables, self.table.normalize())

    def randomize(self):
        self.table.randomize()

    def randomized_copy(self):
        return self.new_instance(self.variables, self.table.randomized_copy())

    def sample(self):
        return self.table.sample()

    def single_sample(self):
        if len(self.table) == 0:
            return []
        else:
            entries_sum = self.table.sum()
            if abs(entries_sum - 1.0) > 0.00001:
                raise Exception(
                    f"Sampled factor is not normalized. Sum of entries is {entries_sum}"
                )
            return discrete_sample(
                self.assignments(), lambda assignment: self.table[assignment]
            )

    def __eq__(self, other):
        """
        Compares factors by checking if they have the same variables and if tables are equal according to ==
        (after appropriate permutation if variables are not in the same order).
        """
        if isinstance(other, TableFactor):
            if self.variables == other.variables:
                return self.table == other.table
            elif set(self.variables) == set(other.variables):
                return self.table == other.get_table_permuted_to_agree_with_table_of(
                    self
                )
            else:
                return False
        else:
            raise Exception(
                f"Comparison of PyTorchFactor to factors other than PyTorchFactor is not implemented. "
                f"Got {type(other)}"
            )

    def get_table_permuted_to_agree_with_table_of(self, other):
        return self.table.permute(permutation_from_to(self.variables, other.variables))

    @property
    def table_factor(self):
        return self

    def __repr__(self):
        return "Factor on (" + join(self.variables) + "): " + repr(self.table)

    def __str__(self):
        return "Factor on (" + join(self.variables) + "): " + str(self.table)



from typing import Any, Dict

from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    generalized_len_of_dict_frame,
)


def dict_data_generator(dictionary: Dict[Any, Any], length, batch_size=100):
    for i in range(0, length, batch_size):
        yield {k: v[i : i + batch_size] for k, v in dictionary.items()}


class DictDataLoader:
    """
    A data loader for datasets organized as a dict, where each key maps to an array-like set of values.
    The data loader returns dicts with the same keys and batches of the array-like values.
    """

    def __init__(self, dictionary: Dict[Any, Any], batch_size=100):
        self.dictionary = dictionary
        self.batch_size = batch_size
        self.length = generalized_len_of_dict_frame(dictionary)
        if batch_size <= 0:
            raise BatchSizeMustBeGreaterThanZero()

    def __iter__(self):
        return dict_data_generator(self.dictionary, self.length, self.batch_size)

    def __len__(self):
        return self.length


class BatchSizeMustBeGreaterThanZero(BaseException):
    def __init__(self):
        super(BatchSizeMustBeGreaterThanZero, self).__init__(
            "Batch size must be greater than zero"
        )



from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.variable.discrete_variable import (
    DiscreteVariable,
)
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.util import util


class MLPFactor(NeuralFactor):
    """A convenience class for a NeuralFactor using an MLP network"""

    @staticmethod
    def make(input_variables, *hidden_layer_sizes_and_output_variable):
        hidden_layer_sizes, output_variable = MLPFactor.validate_arguments(
            input_variables, hidden_layer_sizes_and_output_variable
        )
        neural_net = MLPFactor.make_neural_net(
            input_variables, hidden_layer_sizes, output_variable
        )
        return MLPFactor(neural_net, input_variables, output_variable)

    @staticmethod
    def make_neural_net(input_variables, hidden_layer_sizes, output_variable):
        all_layer_sizes = MLPFactor.compute_mlp_all_layer_sizes_list(
            input_variables, hidden_layer_sizes, output_variable
        )
        neural_net = MLP(*all_layer_sizes)
        return neural_net

    @staticmethod
    def compute_mlp_all_layer_sizes_list(
        input_variables, hidden_layer_sizes, output_variable
    ):
        first_layer_size = len(input_variables)
        last_layer_size = output_variable.cardinality
        all_layer_sizes = [first_layer_size, *hidden_layer_sizes, last_layer_size]
        return all_layer_sizes

    @staticmethod
    def validate_arguments(input_variables, args):
        try:
            hidden_layer_sizes = args[:-1]
            output_variable = args[-1]
            assert (
                all(isinstance(v, DiscreteVariable) for v in input_variables)
                and all(isinstance(i, int) for i in hidden_layer_sizes)
                and isinstance(output_variable, DiscreteVariable)
            )
        except Exception:
            raise Exception(
                f"MLPFactor constructor must receive a list of input variables (all of type {DiscreteVariable.__name__}), "
                f"the sizes of hidden layer (each an argument of the constructor as opposed to being provided in"
                f"a sequence), "
                f"and an output variable (all also of type {DiscreteVariable.__name__}) but got instead "
                f"{input_variables} and {util.join(args)}"
            )
        return hidden_layer_sizes, output_variable



from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import (
    GraphicalModelSGDLearner,
)
from neuralpp.inference.graphical_model.representation.model.model import (
    cross_entropy_for_dataset,
)
from neuralpp.inference.graphical_model.representation.random.random_dataset import (
    generate_dataset,
)
from neuralpp.inference.graphical_model.representation.random.random_model import (
    generate_model,
)
from neuralpp.util.util import (
    assert_equal_up_to_relative_tolerance,
    join,
    try_noisy_test_up_to_n_times,
)


def noisy_test():
    ground_truth_model = generate_model(
        number_of_factors=6, number_of_variables=6, cardinality=3
    )
    print("Model:")
    print(join(ground_truth_model, "\n"))
    print()

    ds = generate_dataset(
        ground_truth_model,
        number_of_sets_of_observed_and_query_variables=50,
        number_of_query_variables=2,
        number_of_observed_variables=2,
        number_of_observations_per_random_set_of_observed_and_query_variables=1,
        datapoints_per_observation=10,
    )
    print()
    print(f"Generated dataset (5 out of {len(ds)} first rows):")
    print(join(ds[:5], "\n"))

    cross_entropy_for_ground_model = cross_entropy_for_dataset(ds, ground_truth_model)
    print()
    print(
        f"Ground truth model loss (cross entropy) for this dataset: {cross_entropy_for_ground_model:.3f}"
    )

    m = [f.randomized_copy() for f in ground_truth_model]

    device = None
    GraphicalModelSGDLearner(m, ds, loss_decrease_tol=1e-2, device=device).learn()

    for f1, f2 in zip(ground_truth_model, m):
        print()
        print(f"Original factor: {f1.normalize()}")
        print(f"Learned  factor: {f2.normalize()}")

    cross_entropy = cross_entropy_for_dataset(ds, m).item()

    print(f"\nGround truth  cross entropy: {cross_entropy_for_ground_model:.3f}")
    print(f"Learned model cross entropy: {cross_entropy:.3f}")

    assert_equal_up_to_relative_tolerance(
        cross_entropy, cross_entropy_for_ground_model, 0.1
    )


def test_learn():
    try_noisy_test_up_to_n_times(noisy_test, n=3)



import random

from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import (
    GraphicalModelSGDLearner,
)
from neuralpp.inference.graphical_model.representation.factor.fixed.fixed_pytorch_factor import (
    FixedPyTorchTableFactor,
)
from neuralpp.inference.graphical_model.representation.factor.neural.MLPFactor import (
    MLPFactor,
)
from neuralpp.inference.graphical_model.representation.model.model import (
    compute_query_probability,
    cross_entropy_for_dataset,
)
from neuralpp.inference.graphical_model.representation.random.random_dataset import (
    generate_dataset_given_observation_and_query_variables,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.util import util
from neuralpp.util.util import try_noisy_test_up_to_n_times


def joint_learning():

    x = IntegerVariable("x", 2)
    y = IntegerVariable("y", 2)
    z = IntegerVariable("z", 2)

    y_is_negation_of_x = FixedPyTorchTableFactor.from_function(
        [x, y], lambda vx, vy: float(vy == 1 - vx)
    )
    noise = 0.9
    z_noisy_equals_y = FixedPyTorchTableFactor.from_function(
        [y, z], lambda vy, vz: noise if vz == vy else (1 - noise)
    )

    ground_truth_model = [y_is_negation_of_x, z_noisy_equals_y]

    dataset = generate_dataset_given_observation_and_query_variables(
        ground_truth_model,
        observed_variables=[x],
        query_variables=[y, z],
        number_of_observations=100,
        datapoints_per_observation=1,
    )

    print("Dataset sample:")
    print(util.join(random.sample(dataset, 20), "\n"))

    y_from_x = MLPFactor.make([x], y)
    learned_model = [y_from_x, z_noisy_equals_y]
    ground_truth_cross_entropy = cross_entropy_for_dataset(
        dataset, ground_truth_model
    ).item()

    print(f"Ground truth model cross entropy: {ground_truth_cross_entropy :.3f}")

    GraphicalModelSGDLearner(learned_model, dataset, loss_decrease_tol=1e-3).learn()

    cross_entropy = cross_entropy_for_dataset(dataset, learned_model).item()

    print(f"Ground truth model cross entropy: {ground_truth_cross_entropy :.3f}")
    print(f"Learned model cross entropy     : {cross_entropy :.3f}")

    for observation_dict, query_dict in random.sample(dataset, 10):
        likelihood = compute_query_probability(
            observation_dict, query_dict, learned_model
        )
        print(
            f"Given {observation_dict}, answer is {query_dict}, learned model assigns it probability {likelihood.item():.2f}"
        )

    expected_cross_entropy_upper_bound = (
        ground_truth_cross_entropy * 1.2
    )  # based on a few runs and observing that that's the usual approximation
    assert (
        cross_entropy <= expected_cross_entropy_upper_bound
    ), f"Learning performance was poorer than expected. Expected upper bound: {expected_cross_entropy_upper_bound:0.3f}, actual: {cross_entropy:.3f}"


def test_joint_learning():
    try_noisy_test_up_to_n_times(joint_learning, n=3)



from neuralpp.inference.graphical_model.representation.frame.dict_data_loader import (
    BatchSizeMustBeGreaterThanZero,
    DictDataLoader,
)
from neuralpp.inference.graphical_model.representation.frame.dict_frame import (
    DictionaryShouldHaveAtLeastOneItem,
    DictionaryValuesShouldAllHaveTheSameLength,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.util.util import check_that_exception_is_thrown


def test_dict_data_loader():

    name = IntegerVariable("name", 10)
    age = IntegerVariable("age", 10)

    dictionary = {name: [1, 2, 3], age: [34, 45, 56]}
    batch_size = 2
    expected_batches = [
        {name: [1, 2], age: [34, 45]},
        {name: [3], age: [56]},
    ]
    run_dict_data_loader_test(dictionary, expected_batches, batch_size)

    dictionary = {name: [1, 2, 3], age: [34, 45, 56]}
    batch_size = 3
    expected_batches = [dictionary]
    run_dict_data_loader_test(dictionary, expected_batches, batch_size)

    dictionary = {name: [1, 2, 3], age: [34, 45, 56]}
    batch_size = 1
    expected_batches = [
        {name: [1], age: [34]},
        {name: [2], age: [45]},
        {name: [3], age: [56]},
    ]
    run_dict_data_loader_test(dictionary, expected_batches, batch_size)

    dictionary = {name: [1, 2, 3], age: [34, 45, 56]}
    batch_size = 0
    expected_batches = [
        {name: [1], age: [34]},
        {name: [2], age: [45]},
        {name: [3], age: [56]},
    ]
    check_that_exception_is_thrown(
        lambda: run_dict_data_loader_test(dictionary, expected_batches, batch_size),
        BatchSizeMustBeGreaterThanZero,
    )

    dictionary = {}
    batch_size = 3
    expected_batches = [
        {name: [1], age: [34]},
        {name: [2], age: [45]},
        {name: [3], age: [56]},
    ]
    check_that_exception_is_thrown(
        lambda: run_dict_data_loader_test(dictionary, expected_batches, batch_size),
        DictionaryShouldHaveAtLeastOneItem,
    )

    dictionary = {name: [1, 2, 3], age: [34]}
    batch_size = 3
    expected_batches = [
        {name: [1], age: [34]},
        {name: [2], age: [45]},
        {name: [3], age: [56]},
    ]
    check_that_exception_is_thrown(
        lambda: run_dict_data_loader_test(dictionary, expected_batches, batch_size),
        DictionaryValuesShouldAllHaveTheSameLength,
    )


def run_dict_data_loader_test(dictionary, expected_batches, batch_size):
    dl = DictDataLoader(dictionary, batch_size=batch_size)
    actual_batches = list(dl)
    assert actual_batches == expected_batches



from neuralpp.inference.graphical_model.representation.factor.neural.neural_factor import (
    NeuralFactor,
)
from neuralpp.inference.graphical_model.variable.integer_variable import IntegerVariable
from neuralpp.inference.neural_net.MLP import MLP
from neuralpp.test.slow_tests.graphical_model.representation.factor.neural.neural_factor_test_util import (
    check_and_show_conditional_distributions,
)
from neuralpp.util.util import try_noisy_test_up_to_n_times


def mlp_neural_factor_generates_distributions():
    x = IntegerVariable("x", 2)
    y = IntegerVariable("y", 4)
    z = IntegerVariable("z", 4)
    input_variables = [x, y]
    output_variable = z
    neural_net = MLP(len(input_variables), 2, 2, output_variable.cardinality)
    neural_factor = NeuralFactor(neural_net, input_variables, output_variable)

    check_and_show_conditional_distributions(neural_factor)


def test_joint_learning():
    try_noisy_test_up_to_n_times(mlp_neural_factor_generates_distributions, n=3)



class FromLogToProbabilitiesAdapter:
    def __init__(self, neural_net):
        self.neural_net = neural_net

    def to(self, device):
        return FromLogToProbabilitiesAdapter(self.neural_net.to(device))

    def __call__(self, input):
        logs = self.neural_net(input)
        return logs.exp()

    def parameters(self):
        return self.neural_net.parameters()

    def randomize(self):
        self.neural_net.randomize()

    def randomized_copy(self):
        return FromLogToProbabilitiesAdapter(self.neural_net.randomized_copy())

    def __repr__(self):
        return f"{self.neural_net}'s probabilities"



from neuralpp.inference.graphical_model.representation.factor.pytorch_table_factor import (
    PyTorchTableFactor,
)


class FixedPyTorchTableFactor(PyTorchTableFactor):

    # Main functionality: overriding pytorch_parameters so empty list so it does not get changed during training

    def pytorch_parameters(self):
        return []

    # Overridden construction

    def __init__(
        self, variables, array_or_table_of_potentials, log_space=True, batch=False
    ):
        super().__init__(
            variables, array_or_table_of_potentials, log_space=log_space, batch=batch
        )

    @staticmethod
    def from_function(variables, function, log_space=True, batch_size=None):
        return PyTorchTableFactor.from_function_and_constructor(
            variables,
            function,
            FixedPyTorchTableFactor,
            log_space=log_space,
            batch_size=batch_size,
        )

    @staticmethod
    def from_predicate(variables, predicate, log_space=True, batch_size=None):
        return FixedPyTorchTableFactor.from_function(
            variables, lambda *args: float(predicate(*args)), log_space, batch_size
        )



# import torch
# from torch.optim import Adam

# from neuralpp.inference.graphical_model.learn.graphical_model_sgd_learner import GraphicalModelSGDLearner
# from neuralpp.inference.graphical_model.variable.discrete_variable import DiscreteVariable
# from neuralpp.inference.graphical_model.variable.tensor_variable import TensorVariable

# from neuralpp.inference.neural_net.ConvNet import ConvNet
# from neuralpp.inference.neural_net.from_log_to_probabilities_adapter import FromLogToProbabilitiesAdapter
# from neuralpp.util.data_loader_from_random_data_point_thunk import data_loader_from_random_data_point_generator
# from neuralpp.util.mnist_util import read_mnist
# from neuralpp.util.util import set_default_tensor_type_and_return_device


# # This script tests whether it is possible for a ConvNet network to learn the digits 0 and 9
# # by negative feedback, namely telling them what they are *not*.
# # More specifically, when the input is the image of a digit other than 0, the epoch_average_loss is
# # the *negative* of the epoch_average_loss we would use to *predict* 0 (analogously for 9).
# # We expect the network to learn what is *not* a 0, and indirectly learn what *is*
# # (analogously for 9).
# # This script is a simplified version of mnist_semi_supervised.py.


# number_of_digits = 10  # actual script will use digits 0..number_of_digits and treat (number_of_digits - 1) as the '9'
#                        # in the experiment's description.
# number_of_batches_per_epoch = 100
# use_a_single_image_per_digit = False
# max_real_mnist_datapoints = None
# try_cuda = True


# def main():

#     neural_net = FromLogToProbabilitiesAdapter(ConvNet())
#     data_loader = data_loader_from_random_data_point_generator(number_of_batches_per_epoch, batch_generator, print=None)

#     global images_by_digits_by_phase  # so they are easily accessible in later functions
#     global next_image_index_by_digit
#     images_by_digits_by_phase = read_mnist(max_real_mnist_datapoints)
#     number_of_training_images = sum([len(images_by_digits_by_phase["train"][d]) for d in range(number_of_digits)])
#     print(f"Loaded {number_of_training_images:,} training images")
#     next_image_index_by_digit = {d: 0 for d in range(number_of_digits)}
#     from_digit_batch_to_image_batch = get_next_real_image_batch_for_digit_batch

#     device = set_default_tensor_type_and_return_device(try_cuda)
#     print(f"Using {device} device")

#     print("\nInitial evaluation:")
#     print_posterior_of(neural_net)

#     print("Learning...")

#     print("\nFinal evaluation:")
#     print_posterior_of(neural_net)


# def batch_generator():
#     pass


# def get_next_real_image_batch_for_digit_batch(digit_batch):
#     images_list = []
#     for d in digit_batch:
#         d = d.item()
#         image = images_by_digits_by_phase["train"][d][next_image_index_by_digit[d]]
#         if use_a_single_image_per_digit:
#             pass  # leave the index at the first position forever
#         else:
#             next_image_index_by_digit[d] += 1
#             if next_image_index_by_digit[d] == len(images_by_digits_by_phase["train"][d]):
#                 next_image_index_by_digit[d] = 0
#         images_list.append(image)
#     images_batch = torch.stack(images_list).to(digit_batch.device)
#     return images_batch


# def print_posterior_of(recognizer, **kwargs):
#     device = kwargs.get('device')
#     for digit in range(number_of_digits):
#         digit_batch = torch.tensor([digit])
#         image_batch = from_digit_batch_to_image_batch(digit_batch)
#         if device is not None:
#             digit_batch = digit_batch.to(device)
#         posterior_probability = recognizer(image_batch)
#         print_posterior(digit, posterior_probability)


# def print_posterior(digit, output_probability):
#     print(f"Prediction for image of {digit}: {output_probability}")



from neuralpp.inference.graphical_model.representation.factor.product_factor import (
    ProductFactor,
)
from neuralpp.inference.graphical_model.variable.variable import Variable


def default_choose_next_to_eliminate(query_variables, product_of_remaining_factors):
    for factor in ProductFactor.factors(product_of_remaining_factors):
        for variable in factor.variables:
            if variable not in query_variables:
                return variable
    return None


class VariableElimination:
    def __init__(self, choose_next_to_eliminate=default_choose_next_to_eliminate):
        self.choose_next_to_eliminate = choose_next_to_eliminate

    def run(self, query, factors):
        """
        Returns a representation equivalent to sum_Others prod factors,
        where Others is all variables but query (which can be a single variable or collection thereof).
        """

        query_variables = [query] if isinstance(query, Variable) else query

        product_of_remaining_factors = ProductFactor(factors)
        while next_variable := self.choose_next_to_eliminate(
            query_variables, product_of_remaining_factors
        ):
            product_of_remaining_factors = product_of_remaining_factors ^ next_variable

        return product_of_remaining_factors.normalize()
